from simpy.core import BoundClass, Environment
from simpy.events import NORMAL, Timeout

from hwt.hdl.value import Value
from hwt.simulator.hdlSimConfig import HdlSimConfig
from hwt.simulator.simModel import mkUpdater, mkArrayUpdater
from hwt.simulator.simulatorCore import HdlProcess
from hwt.simulator.utils import valueHasChanged
from hwt.synthesizer.uniqList import UniqList


def isEvDependentOn(sig, process):
    if sig is None:
        return False
    return process in sig.simFallingSensProcs\
        or process in sig.simRisingSensProcs


class UpdateSet(set):
    """
    Set of updates for signal

    :ivar destination: signal which are updates for
    """

    def __init__(self, destination):
        self.destination = destination


class IoContainer():
    """
    Container for outputs of process
    """
    __slots__ = ["_all_signals"]

    def __init__(self, dstSignalsTuples):
        """
        :param dstSignalsTuples: tuples (name, signal)
        """
        self._all_signals = []
        for name, s in dstSignalsTuples:
            o = UpdateSet(s)
            setattr(self, name, o)
            self._all_signals.append(o)


class HdlSimulator(Environment):
    """
    Circuit simulator with support for external agents

    .. note:: *Signals without driver, constant driver, initial value*
        Every signal is initialized at start with its default value

    .. note:: *Communication between processes*
        Every interprocess signal is marked by synthesizer.
        For each output for every process there is an IO object
        which is container container of updates to signals.
        Every process has (generated) sensitivity-list.
        Process is reevaluated when there is a new value on any signal
        from sensitivity list.

    .. note: *Delta steps*
        Delta step is minimum quantum of changes in simulation, on the begining
        of delta step all read are performed and on the end all writes
        are performed. Writes are causing revalution of HWprocesses
        which are planet into next delta step.
        Delta steps does not update time.
        When there is no process to reevaluate that means thereis nothing to do
        in delta step this delta step is considered
        as last in this time and time is shifted on begining of next event
        by simulator.

    .. note:: *Simulation inputs*
        HWprocess can not contain any blocking statement
        Simulation processes are written in python and can contain anything.
        (Using hdl as main simulator driver is not efficient.
         That is why it is not supported.)

    .. note::
        HWprocesses have lower priority than simulation processes
        this allows simplify logic of all agents.
        When simulation process is executed, HW part did not anything
        in this time, Simulation process can prepare anything for HW part
        (= can write) if simulation process need to read, it has to yield
        simulator.updateComplete event first, process then will be wakened
        after reaction of HW in this time:
        agents are greatly simplified, they just need to yield
        simulator.updateComplete before first read
        and then can not write in this time

    :ivar now: actual simulation time
    :ivar updateComplete: this event is triggered
        when there are not any values to apply in this time
    :ivar valuesToApply: is container of values
        which should be applied in single delta step
    :ivar env: simply environment
    :ivar applyValuesPlaned: flag if there is planed applyValues
        for current values quantum
    :ivar seqProcsToRun: list of event dependent processes
        which should be evaluated after applyValEv
    """
    # updating of combinational signals (wire updates)
    PRIORITY_APPLY_COMB = NORMAL + 1
    # simulation agents waiting for updateComplete event
    PRIORITY_AGENTS_UPDATE_DONE = PRIORITY_APPLY_COMB + 1
    # updateing of event dependent signals (writing in gegisters,rams etc)
    PRIORITY_APPLY_SEQ = PRIORITY_AGENTS_UPDATE_DONE + 1

    process = BoundClass(HdlProcess)
    wait = BoundClass(Timeout)

    def __init__(self, config=None):
        super(HdlSimulator, self).__init__()
        if config is None:
            # default config
            config = HdlSimConfig()

        self.config = config
        self.combUpdateDoneEv = None
        self.applyValEv = None
        self.runSeqProcessesEv = None

        # (signal, value) tuples which should be applied before
        # new round of processes
        #  will be executed
        self.valuesToApply = []
        self.seqProcsToRun = UniqList()
        self.combProcsToRun = UniqList()
        # container of outputs for every process
        self.outputContainers = {}

    def waitOnCombUpdate(self):
        """
        Sim processes can wait on combUpdateDoneEv by:
        yield sim.waitOnCombUpdate()

        Sim process is then woken up when all combinational updates
        are done in this delta step
        """
        cud = self.combUpdateDoneEv
        if cud is None:
            return self.scheduleCombUpdateDoneEv()
        return cud

    def addHwProcToRun(self, trigger, proc):
        # first process in time has to plan executing of apply values on the
        # end of this time
        if self.applyValEv is None:
            # (apply on end of this time to minimalize process reevaluation)
            self.scheduleApplyValues()

        if isEvDependentOn(trigger, proc):
            if self._now == 0:
                return  # pass event dependent on startup
            self.seqProcsToRun.append(proc)
        else:
            self.combProcsToRun.append(proc)

    def _initUnitSignals(self, unit):
        """
        * Inject default values to simulation

        * Instantiate IOs for every process
        """
        for s in unit._cntx.signals:
            v = s.defaultVal.clone()

            # force update all signals to deafut values and propagate it
            s.simUpdateVal(self, mkUpdater(v, False))

        for u in unit._units:
            self._initUnitSignals(u)

        for p in unit._processes:
            self.addHwProcToRun(None, p)

        for p, outputs in unit._outputs.items():
            # name has to be explicit because it may be possible that signal
            # with has this name was replaced by signal from parent/child
            containerNames = list(map(lambda x: x[0], outputs))

            class SpecificIoContainer(IoContainer):
                __slots__ = containerNames

            self.outputContainers[p] = SpecificIoContainer(outputs)

    def __deleteCombUpdateDoneEv(self, ev):
        """
        Callback called on combUpdateDoneEv finished
        """
        self.combUpdateDoneEv = None

    def scheduleCombUpdateDoneEv(self):
        """
        Scheduele combUpdateDoneEv event to let agents know that current
        delta step is ending and values from combinational logic are stable
        """
        assert self.combUpdateDoneEv is None
        cud = self.combUpdateDoneEv = self.event()
        cud.callbacks.append(self.__deleteCombUpdateDoneEv)
        cud._ok = True
        cud._value = None
        self.schedule(cud,
                      priority=self.PRIORITY_AGENTS_UPDATE_DONE)
        return cud

    def scheduleApplyValues(self):
        """
        Apply stashed values to signals
        """
        assert self.applyValEv is None, self.now
        applyVal = self.applyValEv = self.event()
        applyVal._ok = True
        applyVal._value = None
        applyVal.callbacks.append(self.applyValues)

        self.schedule(applyVal,
                      priority=self.PRIORITY_APPLY_COMB)

        if self.runSeqProcessesEv is not None:
            # if runSeqProcessesEv is already scheduled
            return

        assert not self.seqProcsToRun, self.now
        runSeq = self.runSeqProcessesEv = self.event()
        runSeq._ok = True
        runSeq._value = None
        runSeq.callbacks.append(self.runSeqProcesses)

        self.schedule(runSeq, priority=self.PRIORITY_APPLY_SEQ)

    def conflictResolveStrategy(self, actionSet):
        """
        This functions resolves write conflicts for signal

        :param actionSet: set of actions made by process
        """
        invalidate = False
        asLen = len(actionSet)
        # resolve if there is write collision
        if asLen == 0:
            return
        elif asLen == 1:
            res = actionSet.pop()
        else:
            # we are driving signal with two or more different values
            # we have to invalidate result
            res = actionSet.pop()
            invalidate = True

        resLen = len(res)
        if resLen == 3:
            # update for item in array
            val, indexes, isEvDependent = res
            return (mkArrayUpdater(val, indexes, invalidate), isEvDependent)
        else:
            # update for simple signal
            val, isEvDependent = res
            return (mkUpdater(val, invalidate), isEvDependent)

    def runCombProcesses(self):
        """
        Delta step for combinational processes
        """
        for proc in self.combProcsToRun:
            outContainer = self.outputContainers[proc]
            proc(self, outContainer)
            for actionSet in outContainer._all_signals:
                if actionSet:
                    res = self.conflictResolveStrategy(actionSet)
                    # prepare update
                    updater, isEvDependent = res
                    self.valuesToApply.append(
                        (actionSet.destination, updater, isEvDependent, proc))
                    actionSet.clear()
                # else value is latched

        self.combProcsToRun = UniqList()

    def runSeqProcesses(self, ev):
        """
        Delta step for event dependent processes
        """
        updates = []
        for proc in self.seqProcsToRun:
            try:
                outContainer = self.outputContainers[proc]
            except KeyError:
                # processes does not have to have outputs
                outContainer = None

            proc(self, outContainer)

            if outContainer is not None:
                updates.append(outContainer)

        self.seqProcsToRun = UniqList()
        self.runSeqProcessesEv = None

        for cont in updates:
            for actionSet in cont._all_signals:
                if actionSet:
                    v = self.conflictResolveStrategy(actionSet)
                    updater, _ = v
                    actionSet.destination.simUpdateVal(self, updater)
                    actionSet.clear()

    def applyValues(self, ev):
        """
        Perform delta step by writing stacked values to signals
        """
        va = self.valuesToApply
        self.applyValEv = None

        # log if there are items to log
        lav = self.config.logApplyingValues
        if va and lav:
            lav(self, va)
        self.valuesToApply = []

        # apply values to signals, values can overwrite each other
        # but each signal should be driven by only one process and
        # it should resolve value collision
        addSp = self.seqProcsToRun.append
        for s, vUpdater, isEventDependent, comesFrom in va:
            if isEventDependent:
                # now=0 and this was process initialization or async reg
                addSp(comesFrom)
            else:
                # regular combinational process
                s.simUpdateVal(self, vUpdater)

        self.runCombProcesses()

        # processes triggered from simUpdateVal can add new values
        if self.valuesToApply and self.applyValEv is None:
            self.scheduleApplyValues()

    def read(self, sig):
        """
        Read value from signal or interface
        """
        try:
            v = sig._val
        except AttributeError:
            v = sig._sigInside._val

        return v.clone()

    def write(self, val, sig):
        """
        Write value to signal or interface.
        """
        # get target RtlSignal
        try:
            simSensProcs = sig.simSensProcs
        except AttributeError:
            sig = sig._sigInside
            simSensProcs = sig.simSensProcs

        # type cast of input value
        t = sig._dtype

        if isinstance(val, Value):
            v = val.clone()
            v = v._auto_cast(t)
        else:
            v = t.fromPy(val)

        # can not update value in signal directly due singnal proxies
        sig.simUpdateVal(self, lambda curentV: (
            valueHasChanged(curentV, v), v))

        if self.applyValEv is None:
            if not (simSensProcs or
                    sig.simRisingSensProcs or
                    sig.simFallingSensProcs):
                # signal value was changed but there are no sensitive processes
                # to it because of this applyValues is never planed
                # and should be
                self.scheduleApplyValues()
            elif (sig._writeCallbacks or
                  sig._writeCallbacksToEn):
                # signal write did not caused any change on any other signal
                # but there are still simulation agets waiting on
                # updateComplete event
                self.scheduleApplyValues()

    def simUnit(self, synthesisedUnit, time, extraProcesses=[]):
        """
        Run simulation
        """
        beforeSim = self.config.beforeSim
        if beforeSim is not None:
            beforeSim(self, synthesisedUnit)

        for p in extraProcesses:
            self.process(p(self))

        self._initUnitSignals(synthesisedUnit)
        self.run(until=time)



from hwt.hdl.assignment import Assignment
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import isEventDependentOp
from hwt.hdl.portItem import PortItem
from hwt.hdl.statements import IfContainer, SwitchContainer
from hwt.hdl.value import Value
from hwt.pyUtils.arrayQuery import where
from hwt.synthesizer.param import Param
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.exceptions import MultipleDriversExc


def signalHasDriver(sig):
    for _ in walkSignalDrivers(sig):
        return True
    return False


def walkSignalDrivers(sig):
    def assign2Me(ep):
        if isinstance(ep, Assignment):
            return True
        elif isinstance(ep, PortItem) and ep.dst is sig:
            return True
        else:
            return None

    return where(sig.drivers, assign2Me)


def walkAllOriginSignals(sig, discovered=None):
    """
    Walk every signal which has no driver and is used as driver of this signal.
    Goal is walk every generic in static expr. evaluation.
    """
    if discovered is None:
        discovered = set()
    if isinstance(sig, Value):
        return
    if not isinstance(sig, RtlSignalBase):
        raise AssertionError("Expected only instances of signal, got: %s"
                             % (repr(sig)))
    if sig in discovered:
        return
    discovered.add(sig)

    if sig.drivers:
        for obj in sig.drivers:
            if isinstance(obj, Operator):
                for op in obj.operands:
                    if isinstance(op, RtlSignalBase):
                        yield from walkAllOriginSignals(op, discovered=discovered)
            elif isinstance(obj, Assignment):
                yield from walkAllOriginSignals(obj.src, discovered)
            else:
                raise TypeError("walkAllOriginSignals not implemented for %s" % (str(obj)))
    else:
        yield sig


def discoverEventDependency(sig):
    """
    walk signals drivers and yields whose signals which are in some event operator
    """
    try:
        drivers = sig.drivers
    except AttributeError:
        return

    if len(drivers) == 1:
        d = drivers[0]
        if isinstance(d, Operator):
            if isEventDependentOp(d.operator):
                yield d.operands[0]
            else:
                for op in d.operands:
                    yield from discoverEventDependency(op)


class InOutStmProbe():
    """
    Discover input, outputs and sensitivity of given statements

    :ivar inputs: all seen inputs
    :ivar sensitivity: input which are statements sensitive to
        (they have to be revealuated after change on these inputs)
        set of tuples (sensitivity, signal) where sensitivity
        is member of SENSITIVITY enum
    :ivar seen: all signals which were already visited
    :ivar _eventSensFound: flag telling if event dependent sensitivity
        discovered on actual branch and all other inputs
        should be skipped from sensitivity
    """
    def __init__(self):
        self.inputs = set()
        self.sensitivity = set()
        self.seen = set()
        self._eventSensFound = False

    def discover(self, statement):
        self._discover(statement, True)

    def _discover(self, statement, isTop):
        discoverSequence = self._discoverSequence
        walk = self.walkDrivers
        discover = self._discover
        _eventSensFound = self._eventSensFound

        if isinstance(statement, Assignment):
            if statement.indexes:
                discoverSequence(statement.indexes)
            walk(statement.src, self.sensitivity)
        elif isinstance(statement, IfContainer):
            # if true
            discoverSequence(statement.cond)
            for stm in statement.ifTrue:
                discover(stm, True)

            # elifs
            for cond, stms in statement.elIfs:
                discoverSequence(cond)
                for stm in stms:
                    discover(stm, True)
            # else
            for stm in statement.ifFalse:
                discover(stm, True)

        elif isinstance(statement, SwitchContainer):
            walk(statement.switchOn, self.sensitivity)
            for cond, stms in statement.cases:
                # walkDriversInExpr(cond, seenSet)
                for stm in stms:
                    discover(stm, True)

            for stm in statement.default:
                discover(stm, True)
        else:
            raise TypeError(statement)

        if isTop:
            # if event sensitivity was found cancel it
            # because it was found just in this branch
            self._eventSensFound = _eventSensFound

    def _discoverSequence(self, seq):
        """
        Discover sensitivity for list of signals
        """
        casualSensitivity = set()
        walk = self.walkDrivers
        for c in seq:
            walk(c, casualSensitivity)

        # if event dependent sensitivity found do not add other sensitivity
        if not self._eventSensFound:
            self.sensitivity.update(casualSensitivity)

    def walkDrivers(self, expr, casualSensitivity):
        eventSensFound = self._eventSensFound
        if isinstance(expr, (Value, Param)):
            pass
        elif isinstance(expr, RtlSignalBase):
            if expr._const or expr in self.seen:
                pass
            else:
                self.seen.add(expr)

                if not expr.hidden:
                    self.inputs.add(expr)
                    if not eventSensFound:
                        casualSensitivity.add(expr)
                    return

                try:
                    op = expr.singleDriver()
                except MultipleDriversExc:
                    self.inputs.add(expr)
                    if not eventSensFound:
                        casualSensitivity.add(expr)
                    return

                if not isinstance(op, Operator):
                    self.inputs.add(expr)
                    if not eventSensFound:
                        casualSensitivity.add(expr)
                    return

                if isEventDependentOp(op.operator):
                    self.inputs.add(op.operands[0])
                    if self._eventSensFound:
                        assert op in self.sensitivity, "one clock per register"
                    self._eventSensFound = True
                    self.sensitivity.add(op)
                else:
                    # walk source of signal
                    for operand in op.operands:
                        self.walkDrivers(operand, casualSensitivity)
        else:
            raise TypeError(expr)



from hwt.hdl.constants import INTF_DIRECTION


def autoAddAgents(unit):
    """
    Walk all interfaces on unit and instantiate agent for every interface.

    :return: all monitor/driver functions which should be added to simulation
         as processes
    """
    proc = []
    for intf in unit._interfaces:
        if not intf._isExtern:
            continue

        if intf._isInterfaceArray():
            agentCnt = int(intf._asArraySize)
            agents = []
            for i in range(agentCnt):
                _intf = intf[i]
                _intf._initSimAgent()
                assert _intf._ag is not None, intf
                agents.append(_intf._ag)
        else:
            intf._initSimAgent()
            assert intf._ag is not None, intf
            agents = [intf._ag, ]

        if intf._direction == INTF_DIRECTION.MASTER:
            agProcs = list(map(lambda a: a.getMonitors(), agents))
        elif intf._direction == INTF_DIRECTION.SLAVE:
            agProcs = list(map(lambda a: a.getDrivers(), agents))
        else:
            raise NotImplementedError("intf._direction %r for %r" % (
                intf._direction, intf))

        for p in agProcs:
            proc.extend(p)

    return proc


def valuesToInts(values):
    """
    Iterable of values to ints (nonvalid = None)
    """
    res = []
    append = res.append
    for d in values:
        if isinstance(d, int):
            append(d)
        else:
            append(valToInt(d))
    return res


def valToInt(v):
    if v.vldMask == v._dtype.all_mask():
        return v.val
    else:
        return None


def agInts(interface):
    """
    Convert all values which has agent collected in time >=0 to integer array.
    Invalid value will be None.
    """
    return valuesToInts(interface._ag.data)



from time import time

from hwt.pyUtils.arrayQuery import arr_any
from hwt.serializer.ip_packager.busInterface import BusInterface
from hwt.serializer.ip_packager.helpers import appendSpiElem, \
    appendStrElements, mkSpiElm, ns, whereEndsWithExt, whereEndsWithExts
from hwt.serializer.ip_packager.model import Model
from hwt.serializer.ip_packager.otherXmlObjs import VendorExtensions,\
    FileSet, File, Parameter, Value
from hwt.serializer.ip_packager.port import Port
from hwt.synthesizer.interfaceLevel.interfaceUtils.utils import NotSpecified
import xml.etree.ElementTree as etree


vhdl_syn_fileSetName = "xilinx_vhdlsynthesis_view_fileset"
vhdl_sim_fileSetName = "xilinx_vhdlbehavioralsimulation_view_fileset"
tcl_fileSetName = "xilinx_xpgui_view_fileset"


class Component():
    """
    Xilinx xml is element position dependent
    """
    _strValues = ["vendor", "library", "name", "version", "description"]
    # _iterableValues = ["fileSets", "parameters" ]

    def __init__(self):
        self.vendor = ''
        self.library = ''
        self.name = ""
        self.version = "1.0"
        self.busInterfaces = []
        self.model = Model(vhdl_syn_fileSetName,
                           vhdl_sim_fileSetName, tcl_fileSetName)
        self.fileSets = []
        self.description = ""
        self.parameters = []
        self.vendorExtensions = VendorExtensions()
        self._files = []
        self._topUnit = None

    # @classmethod
    # def load(cls, xmlStr):
    #    self = cls()
    #    for prefix, uri in ns.items():
    #        etree.register_namespace(prefix, uri)
    #    self.root = etree.fromstring(xmlStr)
    #    self.name = findS(self.root, "name").text
    #
    #    intf = findS(self.root, "busInterfaces")
    #    self.interfaces = {}
    #    for i in intf:
    #        i = BusInterface.fromElem(i)
    #        if i.name in self.interfaces.keys():
    #            raise Exception("Multiple interfaces with same name")
    #        self.interfaces[i.name] = i
    #    self.model = Model.fromElem(findS(self.root, "model"))
    #
    #    return self

    def _xmlFileSets(self, componentElem):
        def fileSetFromFiles(name, files):
            fileSet = FileSet()
            fileSet.name = name
            for fn in files:
                f = File.fromFileName(fn)
                fileSet.files.append(f)
            return fileSet
        filesets = appendSpiElem(componentElem, "fileSets")
        hdlExtensions = [".vhd", 'v']

        hdl_fs = fileSetFromFiles(vhdl_syn_fileSetName,
                                  whereEndsWithExts(self._files, hdlExtensions))
        hdl_sim_fs = fileSetFromFiles(vhdl_sim_fileSetName,
                                      whereEndsWithExts(self._files, hdlExtensions))
        tclFileSet = fileSetFromFiles(tcl_fileSetName,
                                      whereEndsWithExt(self._files, ".tcl"))
        for fs in [hdl_fs, hdl_sim_fs, tclFileSet]:
            filesets.append(fs.asElem())

    def _xmlParameters(self, compElem):
        parameters = appendSpiElem(compElem, "parameters")
        for p in self.parameters:
            parameters.append(p.asElem())

    def xml(self):
        # Vivado 2015.2 bug - order of all elements is NOT optional
        for prefix, uri in ns.items():
            etree.register_namespace(prefix, uri)
        c = mkSpiElm("component")
        appendStrElements(c, self, self._strValues[:-1])
        if arr_any(self.busInterfaces, lambda intf: hasattr(intf, "_bi")):
            bi = appendSpiElem(c, "busInterfaces")
            for intf in self.busInterfaces:  # for all interfaces which have bus interface class
                if hasattr(intf, "_bi"):
                    bi.append(intf._bi.asElem())

        c.append(self.model.asElem())
        self._xmlFileSets(c)

        appendStrElements(c, self, [self._strValues[-1]])
        self._xmlParameters(c)
        c.append(self.vendorExtensions.asElem(
            self.name + "_v" + self.version, revision=str(int(time()))))

        return c

    def asignTopUnit(self, unit):
        """
        Set hwt unit as template for component
        """
        self._topUnit = unit
        self.name = unit._name
        self.model.addDefaultViews(self._topUnit)
        for p in self._topUnit._entity.ports:
            self.model.ports.append(Port._entPort2CompPort(unit._entity, p))

        for intf in unit._interfaces:
            if intf._isExtern:
                self.busInterfaces.append(intf)

        self.busInterfaces.sort(key=lambda x: x._name)
        for intf in self.busInterfaces:
            biClass = None
            try:
                biClass = intf._getIpCoreIntfClass()
            except NotSpecified:
                pass
            if biClass is not None:
                bi = BusInterface.fromBiClass(intf, biClass)
                intf._bi = bi
                bi.busType.postProcess(
                    self, self._topUnit, self.busInterfaces, intf)

        # generate component parameters
        compNameParam = Parameter()
        compNameParam.name = "Component_Name"
        compNameParam.value = Value()
        v = compNameParam.value
        v.id = "PARAM_VALUE.Component_Name"
        v.resolve = "user"
        v.text = self.name
        self.parameters.append(compNameParam)
        # generic as parameters
        for g in self._topUnit._entity.generics:
            p = Parameter()
            p.name = g.name
            p.value = Value.fromGeneric("PARAM_VALUE.", g, Value.RESOLVE_USER)
            self.parameters.append(p)

        # for bi in self.busInterfaces:
        #    bi.name = trimUnderscores(bi.name)
        #    for p in bi.parameters:
        #        p.name = removeUndescores_witSep(p.name, ".")
        #        p.value.id = removeUndescores_witSep(p.value.id , ".")
        #        p.value.text = removeUndescores_witSep(p.value.text , ".")



from hwt.hdl.hdlObject import HdlObject


class HWProcess(HdlObject):
    """
    Hdl process container

    :ivar name: name used as id in target HDL
    :ivar statements: list of statements in body of process
    :ivar sensitivityList: set of RtlSignals or event operators which
        are describing when should this process be revevaluated
    :ivar inputs: all input signals for this process
    :ivar outputs: all output signals for this process
    :note: HWProcess do not have to be process in target HDL, for example
        simple process which contains only unconditional assignment will
        be rendered just as assignment
    """

    def __init__(self, name, statements, sensitivityList, inputs, outputs):
        self.name = name
        self.statements = statements
        self.sensitivityList = sensitivityList
        self.inputs = inputs
        self.outputs = outputs



import sys

from hwt.serializer.generic.indent import getIndent
from hwt.synthesizer.interfaceLevel.interfaceUtils.proxy import InterfaceProxy
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase


def valueHasChanged(valA, valB):
    return valA.val is not valB.val or valA.vldMask != valB.vldMask


def pprintInterface(intf, prefix="", indent=0, file=sys.stdout):
    """
    Pretty print interface
    """
    try:
        s = intf._sig
    except AttributeError:
        s = ""
    if s is not "":
        s = " " + repr(s)

    file.write("".join([getIndent(indent), prefix, repr(intf._getFullName()),
                        s]))
    file.write("\n")

    for i in intf._interfaces:
        if isinstance(intf, InterfaceProxy):
            assert isinstance(i, InterfaceProxy), (intf, i)
        pprintInterface(i, indent=indent + 1, file=file)

    if intf._arrayElemCache:
        assert len(intf) == len(intf._arrayElemCache)
        for i, p in enumerate(intf):
            pprintInterface(p, prefix="p%d:" % i, indent=indent + 1, file=file)


def pprintAgents(unitOrIntf, indent=0, prefix="", file=sys.stdout):
    if isinstance(unitOrIntf, InterfaceBase):
        ag = unitOrIntf._ag
        arrayElemCache = unitOrIntf._arrayElemCache
    else:
        ag = None
        arrayElemCache = None

    if ag is not None:
        file.write("%s%s%r\n" % (getIndent(indent), prefix, ag))
    elif arrayElemCache:
        file.write("%s%s\n" %
                   (getIndent(indent), prefix + unitOrIntf._name + ":"))

    for i in unitOrIntf._interfaces:
        pprintAgents(i, indent + 1, file=file)

    if arrayElemCache:
        assert len(unitOrIntf) == len(arrayElemCache)
        for i, p in enumerate(unitOrIntf):
            pprintAgents(p, indent + 1, prefix="p%d:" % i, file=file)



from typing import Union, Tuple

from hwt.hdl.types.hdlType import HdlType
from hwt.hdl.types.struct import HStructField


class TransPart(object):
    """
    Container for informations about parts of TransTmpl split on databus words

    :ivar parent: instance of FrameTmpl
    :ivar tmpl: origin template which is this representation of
        (StructField/HdlType instance)
    :ivar isPadding: flag which specifies if this TransaPart is just a padding
    :ivar startOfPart: bit addr of start of this part
    :ivar endOfPart: bit addr of end of this part
    :ivar inFieldOffset: bit offset of this part in parent field
    """

    def __init__(self, parent: 'FrameTmpl',
                 tmpl: Union[HStructField, HdlType],
                 startOfPart: int,
                 endOfPart: int,
                 inFieldOffset: int):
        self.parent = parent
        self.tmpl = tmpl
        self.isPadding = tmpl is None
        self.startOfPart = startOfPart
        self.endOfPart = endOfPart
        self.inFieldOffset = inFieldOffset

    def bit_length(self) -> int:
        """
        :return: bit length of this part
        """
        return self.endOfPart - self.startOfPart

    def getBusWordBitRange(self) -> Tuple[int, int]:
        """
        :return: bit range which contains data of this part on bus data signal
        """
        offset = self.startOfPart % self.parent.wordWidth
        return (offset + self.bit_length(), offset)

    def getFieldBitRange(self) -> Tuple[int, int]:
        """
        :return: bit range which contains data of this part on interface
            of field
        """
        offset = self.inFieldOffset
        return (self.bit_length() + offset, offset)

    def isLastPart(self) -> bool:
        """
        :return: True if this part is last in parts derived from original field
            else False
        """
        return self.tmpl.bitAddrEnd == self.endOfPart

    def __repr__(self):
        return "<TransPart %r, startOfPart:%d, endOfPart:%d>" % (
               self.tmpl, self.startOfPart, self.endOfPart)



def mask(bits):
    return (1 << bits) - 1


def bitField(_from, to):
    """
    _from 0 to 1 -> '1'
    """
    w = to - _from
    return mask(w) << _from


def selectBit(val, bitNo):
    """
    select bit from integer
    """
    return (val >> bitNo) & 1


def selectBitRange(val, bitsStart, bitsLen):
    val >>= bitsStart
    return val & mask(bitsLen)


def clean(val, bitNo):
    return val & ~(1 << bitNo)


def setBit(val, bitNo):
    return val | (1 << bitNo)


def toogle(val, bitNo):
    return val ^ (1 << bitNo)


def setBitRange(val, bitStart, bitsLen, newBits):
    _mask = mask(bitsLen)
    newBits &= _mask

    _mask <<= bitStart
    newBits <<= bitStart

    return (val & ~_mask) | newBits


def bitSetTo(val, bitNo, bitVal):
    if bitVal == 0:
        return clean(val, bitNo)
    elif bitVal == 1:
        return setBit(val, bitNo)
    else:
        raise NotImplementedError()


def align(val, lowerBitCntToAlign):
    val = val >> lowerBitCntToAlign
    return val << lowerBitCntToAlign



from hwt.hdl.operatorDefs import AllOps


class SimModelSerializer_ops():
    opPrecedence = {
        AllOps.NOT: 4,
        AllOps.NEG: 4,
        AllOps.RISING_EDGE: 1,
        AllOps.DIV: 4,
        AllOps.ADD: 5,
        AllOps.SUB: 5,
        AllOps.MUL: 4,
        AllOps.XOR: 9,
        AllOps.EQ: 10,
        AllOps.NEQ: 10,
        AllOps.AND: 10,
        AllOps.OR: 10,
        AllOps.DOWNTO: 1,
        AllOps.GT: 10,
        AllOps.LT: 10,
        AllOps.GE: 10,
        AllOps.LE: 10,
        AllOps.CONCAT: 1,
        AllOps.INDEX: 1,
        AllOps.TERNARY: 1,
        AllOps.CALL: 1,
    }
    _binOps = {
        AllOps.AND: '%s._and__val(%s)',
        AllOps.OR: '%s._or__val(%s)',
        AllOps.XOR: '%s._xor__val(%s)',
        AllOps.CONCAT: "%s._concat__val(%s)",
        AllOps.DIV: '%s._floordiv__val(%s)',
        AllOps.DOWNTO: "SliceVal((%s, %s), SLICE, True)",
        AllOps.EQ: '%s._eq__val(%s)',
        AllOps.GT: '%s._gt__val(%s)',
        AllOps.GE: '%s._ge__val(%s)',
        AllOps.LE: '%s._le__val(%s)',
        AllOps.LT: '%s._lt__val(%s)',
        AllOps.SUB: '%s._sub__val(%s)',
        AllOps.MUL: '%s._mul__val(%s)',
        AllOps.NEQ: '%s._ne__val(%s)',
        AllOps.ADD: '%s._add__val(%s)',
        AllOps.POW: "power(%s, %s)",
    }
    _unaryOps = {
        AllOps.NOT: "(%s)._invert__val()",
        AllOps.RISING_EDGE: "(%s)._onRisingEdge__val(sim.now)",
        AllOps.FALLIGN_EDGE: "(%s)._onFallingEdge__val(sim.now)",
        AllOps.BitsAsSigned: "(%s)._convSign__val(True)",
        AllOps.BitsAsUnsigned: "(%s)._convSign__val(False)",
        AllOps.BitsAsVec: "(%s)._convSign__val(None)",
        AllOps.NEG: "(%s)._neg__val()",
    }

    @classmethod
    def Operator(cls, op, ctx):
        ops = op.operands
        o = op.operator

        op_str = cls._unaryOps.get(o, None)
        if op_str is not None:
            return op_str % (cls._operand(ops[0], o, ctx))

        op_str = cls._binOps.get(o, None)
        if op_str is not None:
            return op_str % (cls._operand(ops[0], o, ctx),
                             cls._operand(ops[1], o, ctx))

        if o == AllOps.INDEX:
            assert len(ops) == 2
            return "(%s)._getitem__val(%s)" % (cls.asHdl(ops[0], ctx),
                                               cls._operand(ops[1], o, ctx))
        elif o == AllOps.TERNARY:
            return "(%s)._ternary__val(%s, %s)" %\
                tuple(map(lambda x: cls.asHdl(x, ctx), ops))
        elif o == AllOps.BitsToInt:
            assert len(ops) == 1
            op = ops[0]
            return "convertSimBits__val(%s, %s, SIM_INT)" % (
                cls.HdlType_bits(op._dtype, ctx),
                cls.asHdl(op, ctx))
        elif o == AllOps.IntToBits:
            assert len(ops) == 1
            resT = op.result._dtype
            return "convertSimInteger__val(%s, %s, %s)" % (
                cls.HdlType(ops[0]._dtype, ctx),
                cls.asHdl(
                    ops[0], ctx),
                cls.HdlType_bits(resT, ctx))
        else:
            raise NotImplementedError(
                "Do not know how to convert %s to simModel" % (o))



from hwt.hdl.typeShortcuts import hInt
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.defs import BIT
from hwt.serializer.ip_packager.exprSerializer import \
    VivadoTclExpressionSerializer
from hwt.serializer.ip_packager.helpers import appendSpiElem, \
    mkSpiElm
from hwt.serializer.vhdl.serializer import VhdlSerializer
from hwt.synthesizer.param import evalParam
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class WireTypeDef():
    _requiredVal = ["typeName"]

    # @classmethod
    # def fromElem(cls, elm):
    #     self = cls()
    #     for s in cls._requiredVal:
    #         setattr(self, s, findS(elm, s).text)
    #     self.viewNameRefs = []
    #     for r in elm.findall("spirit:viewNameRef", ns):
    #         self.viewNameRefs.append(r.text)
    #     return self

    def asElem(self):
        e = mkSpiElm("wireTypeDef")
        for s in self._requiredVal:
            appendSpiElem(e, s).text = getattr(self, s)
        for r in self.viewNameRefs:
            appendSpiElem(e, "viewNameRef").text = r
        return e


class Port():

    # @classmethod
    # def fromElem(cls, elm):
    #     self = cls()
    #     self.name = findS(elm, "name").text
    #     vec = findS(elm, "vector")
    #     if vec is not None:
    #         self.vector = [findS(vec, "left").text, findS(vec, "right").text]
    #     else:
    #         self.vector = None
    #
    #     wire = findS(elm, "wire")
    #     self.direction = findS(wire, "direction").text
    #     self.type = WireTypeDef.fromElem(findS(findS(wire, "wireTypeDefs"),
    #                                            "wiretypedef"))
    #     return self

    @staticmethod
    def _entPort2CompPort(e, p):
        port = Port()
        port.name = p.name
        port.direction = p.direction.name.lower()
        port.type = WireTypeDef()
        t = port.type
        dt = p._dtype

        t.typeName = VhdlSerializer.HdlType(
            dt, VhdlSerializer.getBaseContext())
        try:
            t.typeName = t.typeName[:t.typeName.index('(')]
        except ValueError:
            pass

        if dt == BIT:
            port.vector = False
        elif isinstance(dt, Bits):
            port.vector = [evalParam(dt.width) - 1, hInt(0)]
        t.viewNameRefs = ["xilinx_vhdlsynthesis",
                          "xilinx_vhdlbehavioralsimulation"]
        return port

    def asElem(self):
        e = mkSpiElm("port")
        appendSpiElem(e, "name").text = self.name
        w = appendSpiElem(e, "wire")
        appendSpiElem(w, "direction").text = self.direction
        if self.vector:
            v = appendSpiElem(w, "vector")

            def mkBoundry(name, val):
                if isinstance(val, int):
                    val = hInt(val)
                d = appendSpiElem(v, name)

                d.attrib["spirit:format"] = "long"
                if isinstance(val, RtlSignalBase):
                    # value is simple type and does not contains generic etc...
                    resolve = 'dependent'
                    d.attrib["spirit:dependency"] = "(%s)" %\
                        VivadoTclExpressionSerializer.asHdl(val)
                    d.text = VivadoTclExpressionSerializer.asHdl(
                        val.staticEval())
                else:
                    resolve = "immediate"
                    d.text = VivadoTclExpressionSerializer.asHdl(val, None)
                d.attrib["spirit:resolve"] = resolve
            mkBoundry("left", self.vector[0])
            mkBoundry("right", self.vector[1])
        td = appendSpiElem(w, "wireTypeDefs")
        td.append(self.type.asElem())
        return e



from collections import deque

from hwt.hdl.constants import READ, WRITE, NOP
from hwt.simulator.agentBase import SyncAgentBase
from hwt.simulator.shortcuts import oscilate


class BramPort_withoutClkAgent(SyncAgentBase):
    """
    :ivar requests: list of tuples (request type, address, [write data])
        - used for driver
    :ivar data: list of data in memory, used for monitor
    :ivar mem: if agent is in monitor mode (= is slave) all reads and writes
        are performed on mem object
    """

    def __init__(self, intf):
        super().__init__(intf, allowNoReset=True)

        self.requests = deque()
        self.readPending = False
        self.readed = deque()

        self.mem = {}
        self.requireInit = True

    def doReq(self, sim, req):
        rw = req[0]
        addr = req[1]

        if rw == READ:
            rw = 0
            wdata = None
            self.readPending = True
            if self._debugOutput is not None:
                self._debugOutput.write("%s, after %r read_req: %d\n" % (
                                        self.intf._getFullName(),
                                        sim.now, addr))
        elif rw == WRITE:
            wdata = req[2]
            rw = 1
            if self._debugOutput is not None:
                self._debugOutput.write("%s, after %r write: %d:%d\n" % (
                                        self.intf._getFullName(), sim.now,
                                        addr, wdata))

        else:
            raise NotImplementedError(rw)

        intf = self.intf
        w = sim.write
        w(rw, intf.we)
        w(addr, intf.addr)
        w(wdata, intf.din)

    def onReadReq(self, sim, addr):
        """
        on readReqRecieved in monitor mode
        """
        self.requests.append((READ, addr))

    def onWriteReq(self, sim, addr, data):
        """
        on writeReqRecieved in monitor mode
        """
        self.requests.append((WRITE, addr, data))

    def monitor(self, sim):
        intf = self.intf

        yield sim.waitOnCombUpdate()
        # now we are after clk edge
        if self.notReset(sim):
            en = sim.read(intf.en)
            assert en.vldMask
            if en.val:
                we = sim.read(intf.we)
                assert we.vldMask

                addr = sim.read(intf.addr)
                if we.val:
                    data = sim.read(intf.din)
                    self.onWriteReq(sim, addr, data)
                else:
                    self.onReadReq(sim, addr)

        if self.requests:
            req = self.requests.popleft()
            t = req[0]
            addr = req[1]
            assert addr._isFullVld(), sim.now
            if t == READ:
                sim.write(self.mem[addr.val], intf.dout)
            else:
                assert t == WRITE
                sim.write(None, intf.dout)
                self.mem[addr.val] = req[2]

    def driver(self, sim):
        intf = self.intf
        w = sim.write
        if self.requireInit:
            w(0, intf.en)
            w(0, intf.we)
            self.requireInit = False

#        yield s.wait(2000)
        readPending = self.readPending
        if self.requests and self.notReset(sim):
            req = self.requests.popleft()
            if req is NOP:
                w(0, intf.en)
                w(0, intf.we)
                self.readPending = False
            else:
                self.doReq(sim, req)
                w(1, intf.en)
        else:
            w(0, intf.en)
            w(0, intf.we)
            self.readPending = False

        if readPending:
            yield sim.waitOnCombUpdate()
            # now we are after clk edge
            d = sim.read(intf.dout)
            self.readed.append(d)
            if self._debugOutput is not None:
                self._debugOutput.write("%s, on %r read_data: %d\n" % (
                                        self.intf._getFullName(),
                                        sim.now, d.val))


class BramPortAgent(BramPort_withoutClkAgent):

    def getDrivers(self):
        drivers = super(BramPortAgent, self).getDrivers()
        drivers.append(oscilate(self.intf.clk))
        return drivers



from hwt.hdl.assignment import Assignment
from hwt.hdl.types.typeCast import toHVal
from hwt.hdl.value import Value
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.mainBases import RtlMemoryBase
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


class RtlSyncSignal(RtlMemoryBase, RtlSignal):
    """
    Syntax sugar,
    every write is made to next signal, "next" is assigned
    to main signal on every clock rising edge
    """

    def __init__(self, ctx, name, var_type, defaultVal=None):
        """
        :param ctx: context in which is sig. created (instance of RtlNetlist)
        :param name: suggested name
        :param var_type: type of signal
        :param defaultVal: default value for signal (used as def. val in hdl and for reset)
        """
        super().__init__(ctx, name, var_type, defaultVal)
        self.next = RtlSignal(ctx, name + "_next", var_type,
                              nopVal=self, useNopVal=True)

    def __call__(self, source):
        """
        assign to signal which is next value of this register

        :return: list of assignments
        """
        if isinstance(source, InterfaceBase):
            source = source._sig

        if source is None:
            source = self._dtype.fromPy(None)
        else:
            source = toHVal(source)
            source = source._auto_cast(self._dtype)

        a = Assignment(source, self.next)

        self.next.drivers.append(a)
        if not isinstance(source, Value):
            source.endpoints.append(a)
        return [a]

    def __repr__(self):
        from hwt.serializer.hwt.serializer import HwtSerializer
        ctx = HwtSerializer.getBaseContext()
        return HwtSerializer.SignalItem(self, ctx)



from operator import floordiv, add, sub, inv, mod, mul, ne, and_, or_, \
    xor, gt, ge, lt, le, getitem, neg

from hwt.hdl.constants import SENSITIVITY
from hwt.hdl.types.defs import INT
from hwt.hdl.value import Value


class OpDefinition():
    """
    Operator definition

    :ivar id: name of operator
    :ivar _evalFn: function which evaluates operands
    """

    def __init__(self, evalFn):
        self.id = None  # assigned automatically in AllOps
        self._evalFn = evalFn

    def __eq__(self, other):
        return type(self) == type(other) and self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def eval(self, operator, simulator=None):
        """Load all operands and process them by self._evalFn"""
        def getVal(v):
            while not isinstance(v, Value):
                v = v._val

            return v

        operands = list(map(getVal, operator.operands))

        if isEventDependentOp(operator.operator):
            operands.append(simulator.now)
        elif operator.operator == AllOps.IntToBits:
            operands.append(operator.result._dtype)

        return self._evalFn(*operands)

    def __repr__(self):
        return "<OpDefinition %s>" % (self.id)


def isEventDependentOp(operator):
    return operator in (AllOps.RISING_EDGE, AllOps.FALLIGN_EDGE)


def onRisingEdgeFn(a, now):
    return a._onRisingEdge(now)


def onFallingEdgeFn(a, now):
    return a._onFallingEdge(now)


def dotOpFn(a, name):
    return getattr(a, name)


# [TODO] downto / to are relict of vhdl and should be replaced with slice
def downtoFn(a, b):
    return a._downto(b)


def toFn(a, b):
    return a._to(b)


def concatFn(a, b):
    return a._concat(b)


def power(base, exp):
    return base._pow(exp)


def eqFn(a, b):
    return a._eq(b)


def ternaryFn(a, b, c):
    return a._ternary(b, c)


def callFn(fn, *operands, **kwargs):
    return fn(*operands, **kwargs)


def bitsToIntFn(a):
    return a._auto_cast(INT)


def intToBitsFn(a, t):
    return a._auto_cast(t)


def bitsAsSignedFn(a):
    return a._signed()


def bitsAsUnsignedFn(a):
    return a._unsigned()


def bitsAsVec(a):
    return a._vec()


class AllOps():
    """
    :attention: Remember that and operator "and" is & and "or" is \|, "and"
        and "or" can not be used because they can not be overloaded
    :attention: These are operators of internal AST,
        they are not equal to verilog or vhdl operators
    """
    RISING_EDGE = OpDefinition(onRisingEdgeFn)  # unnecessary
    FALLIGN_EDGE = OpDefinition(onFallingEdgeFn)  # unnecessary

    NEG = OpDefinition(neg)
    DIV = OpDefinition(floordiv)
    ADD = OpDefinition(add)
    SUB = OpDefinition(sub)
    POW = OpDefinition(power)
    UN_MINUS = OpDefinition(inv)
    MOD = OpDefinition(mod)
    MUL = OpDefinition(mul)

    NOT = OpDefinition(inv)
    XOR = OpDefinition(xor)
    AND = OpDefinition(and_)
    OR = OpDefinition(or_)

    DOT = OpDefinition(dotOpFn)
    DOWNTO = OpDefinition(downtoFn)
    TO = OpDefinition(toFn)
    CONCAT = OpDefinition(concatFn)

    EQ = OpDefinition(eqFn)
    NEQ = OpDefinition(ne)
    GT = OpDefinition(gt)
    GE = OpDefinition(ge)
    LT = OpDefinition(lt)
    LE = OpDefinition(le)

    INDEX = OpDefinition(getitem)
    TERNARY = OpDefinition(ternaryFn)
    CALL = OpDefinition(callFn)

    BitsToInt = OpDefinition(bitsToIntFn)
    IntToBits = OpDefinition(intToBitsFn)

    BitsAsSigned = OpDefinition(bitsAsSignedFn)
    BitsAsUnsigned = OpDefinition(bitsAsUnsignedFn)
    BitsAsVec = OpDefinition(bitsAsVec)


for a in dir(AllOps):
    o = getattr(AllOps, a)
    if isinstance(o, OpDefinition):
        o.id = a


def sensitivityByOp(op):
    """
    get sensitivity type for operator
    """
    if op == AllOps.RISING_EDGE:
        return SENSITIVITY.RISING
    elif op == AllOps.FALLIGN_EDGE:
        return SENSITIVITY.FALLING
    else:
        raise TypeError()



from hwt.hdl.value import Value, areValues
from hwt.hdl.types.defs import BOOL
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps

BoolVal = BOOL.getValueCls()


class HEnumVal(Value):
    @classmethod
    def fromPy(cls, val, typeObj, vldMask=None):
        """
        :param val: value of python type bool or None
        :param typeObj: instance of HEnum
        :param vldMask: if is None validity is resolved from val
            if is 0 value is invalidated
            if is 1 value has to be valid
        """
        if val is None:
            assert vldMask is None or vldMask == 0
            valid = False
            val = typeObj._allValues[0]
        else:
            if vldMask is None or vldMask == 1:
                assert isinstance(val, str)
                valid = True
            else:
                valid = False
                val = None

        return cls(val, typeObj, valid)

    def _eq__val(self, other):
        eq = self.val == other.val \
            and self.vldMask == other.vldMask == 1

        vldMask = int(self.vldMask == other.vldMask == 1)
        updateTime = max(self.updateTime, other.updateTime)
        return BoolVal(eq, BOOL, vldMask, updateTime)

    def _eq(self, other):
        assert self._dtype is other._dtype

        if areValues(self, other):
            return self._eq__val(other)
        else:
            return Operator.withRes(AllOps.EQ, [self, other], BOOL)

    def _ne__val(self, other):
        neq = self.val != other.val \
            and self.vldMask == other.vldMask == 1

        vldMask = int(self.vldMask == other.vldMask == 1)
        updateTime = max(self.updateTime, other.updateTime)
        return BoolVal(neq, BOOL, vldMask, updateTime)

    def __ne__(self, other):
        assert self._dtype is other._dtype

        if areValues(self, other):
            return self._ne__val(other)
        else:
            return Operator.withRes(AllOps.NEQ, [self, other], BOOL)



from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.bool import HBool
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.integer import Integer
from hwt.hdl.types.slice import Slice
from hwt.hdl.types.string import String
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase


class GenericSerializer_Value():
    @classmethod
    def Value(cls, val, ctx):
        """
        :param dst: is signal connected with value
        :param val: value object, can be instance of Signal or Value
        """
        t = val._dtype

        if isinstance(val, RtlSignalBase):
            return cls.SignalItem(val, ctx)

        # try to extract value as constant
        try:
            consGetter = ctx.constCache.getConstName
        except AttributeError:
            consGetter = None

        if consGetter and not isinstance(t, HEnum):
            return "self." + consGetter(val)

        if isinstance(t, Slice):
            return cls.Slice_valAsHdl(t, val, ctx)
        elif isinstance(t, HArray):
            return cls.HArrayValAsHdl(t, val, ctx)
        elif isinstance(t, Bits):
            return cls.Bits_valAsHdl(t, val, ctx)
        elif isinstance(t, HBool):
            return cls.Bool_valAsHdl(t, val, ctx)
        elif isinstance(t, HEnum):
            return cls.HEnumValAsHdl(t, val, ctx)
        elif isinstance(t, Integer):
            return cls.Integer_valAsHdl(t, val, ctx)
        elif isinstance(t, String):
            return cls.String_valAsHdl(t, val, ctx)
        else:
            raise Exception("can not resolve value serialization for %r"
                            % (val))

    @classmethod
    def Integer_valAsHdl(cls, dtype, val, ctx):
        return str(int(val.val))

    @classmethod
    def Bits_valAsHdl(cls, dtype, val, ctx):
        w = dtype.bit_length()
        if dtype.signed is None:
            if dtype.forceVector or w > 1:
                return cls.BitString(val.val, w, val.vldMask)
            else:
                return cls.BitLiteral(val.val, val.vldMask)
        elif dtype.signed:
            return cls.SignedBitString(val.val, w, dtype.forceVector,
                                       val.vldMask)
        else:
            return cls.UnsignedBitString(val.val, w, dtype.forceVector,
                                         val.vldMask)



from hwt.hdl.constants import DIRECTION, INTF_DIRECTION
from hwt.hdl.types.typeCast import toHVal
from hwt.synthesizer.exceptions import IntfLvlConfErr
from hwt.synthesizer.interfaceLevel.interfaceUtils.array import InterfaceArray
from hwt.synthesizer.interfaceLevel.interfaceUtils.directionFns import \
    InterfaceDirectionFns
from hwt.synthesizer.interfaceLevel.interfaceUtils.implDependent import\
    InterfaceceImplDependentFns
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.interfaceLevel.propDeclrCollector import\
    PropDeclrCollector
from hwt.synthesizer.param import Param
from hwt.synthesizer.vectorUtils import fitTo


def _defaultUpdater(self, onParentName, p):
    self._replaceParam(onParentName, p)


class Interface(InterfaceBase, InterfaceceImplDependentFns, InterfaceArray,
                PropDeclrCollector, InterfaceDirectionFns):
    """
    Base class for all interfaces in interface synthesizer

    :cvar _NAME_SEPARATOR: separator for nested interface names
    :ivar _params: [] of parameter
    :ivar _interfaces: [] sub interfaces
    :ivar _name: name assigned during synthesis
    :ivar _parent: parent object (Unit or Interface instance)
    :ivar _isExtern: If true synthesizer sets it as external port of unit
    :ivar _associatedClk: clock Signal (interface) associated with
        this interface if is none simulation agent try to search it on parent
    :ivar _associatedRst: rst(_n) Signal (interface) associated
        with this interface if is none simulation agent try to search
        it on parent

    :note: only interfaces without _interfaces have

    :ivar _sig: rtl level signal instance
    :ivar _sigInside: _sig after toRtl conversion is made
        (after toRtl conversion _sig is signal for parent unit
        and _sigInside is signal in original unit, this separates process
        of translating units)
    :ivar _boundedEntityPort: entityPort for which was this interface created
    :ivar _boundedSigLvlUnit: RTL unit for which was this interface created


    Agenda of direction

    :ivar _masterDir: specifies which direction has this interface at master
    :ivar _direction: means actual direction of this interface resolved
        by its drivers
    :ivar _cntx: rtl netlist context of all signals and params
        on this interface after interface is registered on parent _cntx
        is merged

    Agenda of simulations

    :ivar _ag: agent object connected to this interface
        (initialized only before simultion)
    """

    _NAME_SEPARATOR = "_"

    def __init__(self, masterDir=DIRECTION.OUT, asArraySize=None,
                 loadConfig=True):
        """
        This constructor is called when constructing new interface,
        it is usually done manually while creating Unit or
        automatically while extracting interfaces from UnitWithSoure

        :param masterDir: direction which this interface should have for master
        :param multiplyedBy: this can be instance of integer or Param,
            this mean the interface is array of the interfaces
            where multiplyedBy is the size
        :param loadConfig: do load config in __init__
        """
        self._setAttrListener = None
        self._associatedClk = None
        self._associatedRst = None
        self._parent = None

        super().__init__()
        if asArraySize is not None:
            asArraySize = toHVal(asArraySize)
            assert int(asArraySize) > 0
            self._widthMultiplier = asArraySize
        else:
            self._widthMultiplier = None

        self._asArraySize = asArraySize

        self._masterDir = masterDir
        self._direction = INTF_DIRECTION.UNKNOWN

        self._cntx = None

        if loadConfig:
            self._loadConfig()

        # flags for better design error detection
        self._isExtern = False
        self._isAccessible = True
        self._dirLocked = False
        self._ag = None

    def __call__(self, other):
        """
        :attention: it is not call of function it is operator of assignment
        """
        return self._connectTo(other)

    def _loadDeclarations(self):
        """
        load declaratoins from _declr method
        This function is called first for parent and then for children
        """
        if not hasattr(self, "_interfaces"):
            self._interfaces = []
        self._setAttrListener = self._declrCollector
        self._declr()
        self._setAttrListener = None

        for i in self._interfaces:
            # inherit _asArraySize and update dtype on physical interfaces
            w = i._widthMultiplier
            if self._widthMultiplier is not None:
                if w is None:
                    w = self._widthMultiplier
                else:
                    w = w * self._widthMultiplier

            i._widthMultiplier = w
            i._isExtern = self._isExtern
            i._loadDeclarations()

        # apply multiplier at dtype of signals
        if not self._interfaces and self._widthMultiplier is not None:
            self._injectMultiplerToDtype()

        if self._isInterfaceArray():
            self._initArrayItems()

        for p in self._params:
            p.setReadOnly()

    def _clean(self, rmConnetions=True, lockNonExternal=True):
        """
        Remove all signals from this interface (used after unit is synthesized
        and its parent is connecting its interface to this unit)
        """

        if self._interfaces:
            for i in self._interfaces:
                i._clean(rmConnetions=rmConnetions,
                         lockNonExternal=lockNonExternal)
        else:
            self._sigInside = self._sig
            del self._sig

        self._dirLocked = False
        if lockNonExternal and not self._isExtern:
            self._isAccessible = False  # [TODO] mv to signal lock

        if self._isInterfaceArray():
            for e in self._arrayElemCache:
                e._clean(rmConnetions=rmConnetions,
                         lockNonExternal=lockNonExternal)

    def _connectToIter(self, master, exclude=None, fit=False):
        if exclude and (self in exclude or master in exclude):
            return

        if self._interfaces:
            for ifc in self._interfaces:
                if exclude and ifc in exclude:
                    continue
                mIfc = getattr(master, ifc._name)

                if exclude and mIfc in exclude:
                    continue

                if mIfc._masterDir == DIRECTION.OUT:
                    if ifc._masterDir != mIfc._masterDir:
                        raise IntfLvlConfErr(
                            "Invalid connection %r <= %r" % (ifc, mIfc))

                    yield from ifc._connectTo(mIfc,
                                              exclude=exclude,
                                              fit=fit)
                else:
                    if ifc._masterDir != mIfc._masterDir:
                        raise IntfLvlConfErr(
                            "Invalid connection %r <= %r" % (mIfc, ifc))

                    yield from mIfc._connectTo(ifc,
                                               exclude=exclude,
                                               fit=fit)
        else:
            dstSig = toHVal(self)
            srcSig = toHVal(master)

            if fit:
                srcSig = fitTo(srcSig, dstSig)

            yield dstSig(srcSig)

    def _signalsForInterface(self, context, prefix='', typeTransform=None):
        """
        generate _sig for each interface which has no subinterface
        if already has _sig return it instead

        :param context: instance of RtlNetlist where signals should be created
        :param prefix: name prefix for created signals
        :param typeTransform: optional function (type) returns modified type
            for signal
        """
        sigs = []
        if self._interfaces:
            for intf in self._interfaces:
                sigs.extend(
                    intf._signalsForInterface(context, prefix,
                                              typeTransform=typeTransform))
        else:
            if hasattr(self, '_sig'):
                sigs = [self._sig]
            else:
                t = self._dtype
                if typeTransform is not None:
                    t = typeTransform(t)

                s = context.sig(prefix + self._getPhysicalName(), t)
                s._interface = self
                self._sig = s

                if hasattr(self, '_boundedEntityPort'):
                    self._boundedEntityPort.connectSig(self._sig)
                sigs = [s]

        if self._asArraySize is not None:
            for elemIntf in self._arrayElemCache:
                elemIntf._signalsForInterface(context)
                # they are not in sigs because they are not main signals

        return sigs

    def _getPhysicalName(self):
        """Get name in HDL """
        if hasattr(self, "_boundedEntityPort"):
            return self._boundedEntityPort.name
        else:
            return self._getFullName().replace('.', self._NAME_SEPARATOR)

    def _getFullName(self):
        """get all name hierarchy separated by '.' """
        name = ""
        tmp = self
        while isinstance(tmp, InterfaceBase):
            if hasattr(tmp, "_name"):
                n = tmp._name
            else:
                n = ''
            if name == '':
                name = n
            else:
                name = n + '.' + name
            if hasattr(tmp, "_parent"):
                tmp = tmp._parent
            else:
                tmp = None
        return name

    def _replaceParam(self, pName, newP):
        """
        Replace parameter on this interface (in configuration stage)

        :ivar pName: actual name of param on me
        :ivar newP: new Param instance by which should be old replaced
        """
        p = getattr(self, pName)
        i = self._params.index(p)
        assert i > -1
        self._params[i] = newP
        del p._scopes[self]  # remove reference from old param
        newP._registerScope(pName, self)
        setattr(self, pName, newP)

    def _updateParamsFrom(self, otherObj, updater=_defaultUpdater,
                          exclude=None):
        """
        update all parameters which are defined on self from otherObj

        :param exclude: iterable of parameter on other object
            which should be excluded
        """
        excluded = set()
        if exclude is not None:
            exclude = set(exclude)

        for parentP in otherObj._params:
            if exclude and parentP in exclude:
                excluded.add(parentP)
                continue
            _, onParentName = parentP._scopes[otherObj]
            try:
                myP = getattr(self, onParentName)
                if not isinstance(myP, Param):
                    continue
            except AttributeError:
                continue

            updater(self, onParentName, parentP)

        if exclude is not None:
            assert excluded == exclude

    def _bit_length(self):
        """Sum of all width of interfaces in this interface"""
        try:
            interfaces = self._interfaces
        except AttributeError:
            interfaces = None

        if interfaces is None:
            # not loaded interface
            _intf = self._clone()
            _intf._loadDeclarations()
            interfaces = _intf._interfaces

        if interfaces:
            w = 0
            for i in interfaces:
                w += i._bit_length()
            return w
        else:
            return self._dtype.bit_length()

    def _connectTo(self, master, exclude=None, fit=False):
        """
        connect to another interface interface (on rtl level)
        works like self <= master in VHDL
        """
        return list(self._connectToIter(master, exclude, fit))

    def __repr__(self):
        s = [self.__class__.__name__]
        s.append("name=%s" % self._getFullName())
        if hasattr(self, '_width'):
            s.append("_width=%s" % str(self._width))
        if hasattr(self, '_masterDir'):
            s.append("_masterDir=%s" % str(self._masterDir))
        return "<%s>" % (', '.join(s))



from hwt.hdl.constants import DIRECTION


class PortMap():
    def __init__(self, signal, portItem):
        self.sig = signal
        self.portItem = portItem

    @classmethod
    def fromPortItem(cls, portItem):
        d = portItem.direction
        if d == DIRECTION.IN:
            sig = portItem.src
        elif d == DIRECTION.OUT:
            sig = portItem.dst
        else:
            raise NotImplementedError()

        assert sig is not None

        return cls(sig, portItem)

    def __repr__(self):
        return "<PortMap %s => %s>" % (self.portItem.name, self.sig.name)



from hwt.hdl.operator import Operator
from hwt.hdl.portItem import PortItem
from hwt.pyUtils.arrayQuery import arr_any
from hwt.serializer.generic.constants import SIGNAL_TYPE


def systemCTypeOfSig(signalItem):
    """
    Check if is register or wire
    """
    if signalItem._const or\
       arr_any(signalItem.drivers,
               lambda d: not isinstance(d, (PortItem, Operator))
               and d.isEventDependent):

        return SIGNAL_TYPE.REG
    else:
        return SIGNAL_TYPE.WIRE





class TclObj():
    pass


class GuiParam(TclObj):
    def __init__(self, name, parent):
        self.name = name
        self.parent = parent

    def __str__(self):
        return 'ipgui::add_param $IPINST -name "%s" -parent ${%s}' % (
            self.name, self.parent.name)


class GuiPage(TclObj):
    __slots__ = ["name", 'parent']

    def __init__(self, name):
        self.name = name
        self.elements = []

    def param(self, name):
        p = GuiParam(name, self)
        self.elements.append(p)
        return p

    def __str__(self):
        params = []
        for n in self.__slots__:
            p = getattr(self, n, None)
            if p:
                if issubclass(p.__class__, TclObj):
                    s = '-%s ${%s}' % (n, p.name)
                else:
                    s = '-%s "%s"' % (n, p)
                params.append(s)
        page = 'set %s [ipgui::add_page $IPINST %s]' % (
            self.name, ' '.join(params))
        elements = '\n'.join(map(lambda x: str(x), self.elements))
        return '\n'.join([page, elements])


class GuiBuilder():

    def __init__(self):
        self.elements = []

    def page(self, name):
        p = GuiPage(name)
        self.elements.append(p)
        return p

    def asTcl(self):
        init_gui = TclFn("init_gui", ["IPINST"], self.elements)
        return str(init_gui)


class TclFn(TclObj):

    def __init__(self, name, params, body):
        self.name = name
        self.params = params
        self.body = body

    def __str__(self):
        return "proc %s { %s } {\n %s \n}" % (
            self.name,
            ' '.join(
                map(lambda x: str(x), self.params)),
            '\n'.join(map(lambda x: str(x), self.body)))


def setParamOnModel(propName):
    return ("set_property value [get_property value ${PARAM_VALUE.%s}]"
            " ${MODELPARAM_VALUE.%s}") % (propName, propName)


def paramManipulatorFns(paramName):
    pv = 'PARAM_VALUE.' + paramName
    yield TclFn("update_PARAM_VALUE." + paramName, [pv], [])
    yield TclFn("validate_PARAM_VALUE." + paramName, [pv], ["return true"])
    yield TclFn("update_MODELPARAM_VALUE." + paramName,
                ["MODELPARAM_VALUE." + paramName, pv],
                [setParamOnModel(paramName)])



from collections import deque

from hwt.hdl.constants import NOP
from hwt.simulator.agentBase import SyncAgentBase


class HandshakedAgent(SyncAgentBase):
    """
    Simulation/verification agent for :class:`hwt.interfaces.std.Handshaked`
    interface there is onMonitorReady(simulator)
    and onDriverWirteAck(simulator) unimplemented method
    which can be used for interfaces with bi-directional data streams

    :attention: requires clk and rst/rstn signal
        (if you do not have any create simulation wrapper with it)
    """

    def __init__(self, intf):
        super().__init__(intf)
        self.actualData = NOP
        self.data = deque()
        # these signals are extracted like this to make
        # agent more configurable
        self._rd = self.getRd()
        self._vld = self.getVld()

        # tmp variables to keep track of last send values to simulation
        self._lastWritten = None
        self._lastRd = None
        self._lastVld = None

    def setEnable_asDriver(self, en, sim):
        super(HandshakedAgent, self).setEnable_asDriver(en, sim)
        if not en:
            self.wrVld(sim.write, 0)
            self._lastVld = 0

    def setEnable_asMonitor(self, en, sim):
        super(HandshakedAgent, self).setEnable_asMonitor(en, sim)
        if not en:
            self.wrRd(sim.write, 0)
            self._lastRd = 0

    def getRd(self):
        """get "ready" signal"""
        return self.intf.rd._sigInside

    def isRd(self, readFn):
        """
        get value of "ready" signal
        """
        return readFn(self._rd)

    def wrRd(self, wrFn, val):
        wrFn(val, self._rd)

    def getVld(self):
        """get "valid" signal"""
        return self.intf.vld._sigInside

    def isVld(self, readFn):
        """
        get value of "valid" signal, override f.e. when you
        need to use signal with reversed polarity
        """
        return readFn(self._vld)

    def wrVld(self, wrFn, val):
        wrFn(val, self._vld)

    def monitor(self, sim):
        """
        Collect data from interface
        """
        r = sim.read
        if self.notReset(sim):
            # update rd signal only if required
            if self._lastRd is not 1:
                self.wrRd(sim.write, 1)
                self._lastRd = 1

                # try to run onMonitorReady if there is any
                try:
                    onMonitorReady = self.onMonitorReady
                except AttributeError:
                    onMonitorReady = None

                if onMonitorReady is not None:
                    onMonitorReady(sim)

            # wait for response of master
            yield sim.waitOnCombUpdate()
            vld = self.isVld(r)
            assert vld.vldMask, (
                "valid signal for interface %r is in invalid state,"
                " this would cause desynchronization, %d") % (
                    self.intf, sim.now)

            if vld.val:
                # master responded with positive ack, do read data
                d = self.doRead(sim)
                if self._debugOutput is not None:
                    self._debugOutput.write(
                        "%s, read, %d: %r\n" % (
                            self.intf._getFullName(), sim.now, d))
                self.data.append(d)
        else:
            if self._lastRd is not 0:
                # can not receive, say it to masters
                self.wrRd(sim.write, 0)
                self._lastRd = 0

    def doRead(self, sim):
        """extract data from interface"""
        return sim.read(self.intf.data)

    def doWrite(self, sim, data):
        """write data to interface"""
        sim.write(data, self.intf.data)

    def checkIfRdWillBeValid(self, sim):
        yield sim.waitOnCombUpdate()
        rd = self.isRd(sim.read)
        assert rd.vldMask, (
            "ready signal for interface %r is in invalid state,"
            " this would cause desynchronization, %d") % (
                self.intf, sim.now)

    def driver(self, sim):
        """
        Push data to interface

        set vld high and wait on rd in high then pass new data
        """
        r = sim.read

        # pop new data if there are not any pending
        if self.actualData is NOP and self.data:
            self.actualData = self.data.popleft()

        doSend = self.actualData is not NOP

        # update data on signals if is required
        if self.actualData is not self._lastWritten:
            if doSend:
                self.doWrite(sim, self.actualData)
            else:
                self.doWrite(sim, None)
            self._lastWritten = self.actualData

        en = self.notReset(sim)
        vld = int(en and doSend)
        if self._lastVld is not vld:
            self.wrVld(sim.write, vld)
            self._lastVld = vld

        if not self._enabled:
            # we can not check rd it in this function because we can not wait
            # because we can be reactivated in this same time
            sim.process(self.checkIfRdWillBeValid(sim))
            return

        # wait of response of slave
        yield sim.waitOnCombUpdate()

        rd = self.isRd(r)
        assert rd.vldMask, (
            "ready signal for interface %r is in invalid state,"
            " this would cause desynchronization, %d") % (
                self.intf, sim.now)
        if not vld:
            return

        if rd.val:
            # slave did read data, take new one
            if self._debugOutput is not None:
                self._debugOutput.write("%s, wrote, %d: %r\n" % (
                    self.intf._getFullName(),
                    sim.now,
                    self.actualData))

            # pop new data, because actual was read by slave
            if self.data:
                self.actualData = self.data.popleft()
            else:
                self.actualData = NOP

            # try to run onDriverWirteAck if there is any
            try:
                onDriverWriteAck = self.onDriverWirteAck
            except AttributeError:
                onDriverWriteAck = None
            if onDriverWriteAck is not None:
                onDriverWriteAck(sim)


class HandshakeSyncAgent(HandshakedAgent):
    """
    Simulation/verification agent for HandshakedSycn interface

    :attention: there is no data channel on this interface
        it is synchronization only and it actually does not have
        any meaningful data collected data in monitor
        mode are just values of simulation time when item was collected
    """

    def doWrite(self, sim, data):
        pass

    def doRead(self, sim):
        return sim.now



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from hwt.hdl.types.typeCast import toHVal
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.rtlSignal import RtlSignal


class Param(RtlSignal):
    """
    Class used in same way as generics in VHDL, it is wrapper around the value
    """

    def __init__(self, initval):
        initval = toHVal(initval)
        super(Param, self).__init__(None, None, initval._dtype,
                                    defaultVal=initval)
        self._val = initval
        self.replacedWith = None
        self._parent = None
        self.__isReadOnly = False
        # unit: (ctx, name)
        self._scopes = {}
        self._const = True

    def _registerScope(self, name, unit):
        self._scopes[unit] = (unit._cntx, name)

    def getName(self, where):
        return self._scopes[where][1]

    def setReadOnly(self):
        self.__isReadOnly = True

    def get(self):
        if self.replacedWith is not None:
            raise Exception("Trying to read param '%r' which is already"
                            " replaced by '%r'" % (self, self.replacedWith))
        return self._val

    def set(self, val):
        """
        set value of this param
        """
        assert not self.__isReadOnly, \
            ("This parameter(%s) was locked"
             " and now it can not be changed" % self.name)
        assert self.replacedWith is None, \
            ("This param was replaced with new one and this "
             "should not exists")

        val = toHVal(val)
        self.defaultVal = val
        self._val = val.staticEval()

    def __repr__(self):
        val = "InvalidVal"
        try:
            val = self.get()
        except Exception:
            pass

        try:
            name = self.name
        except AttributeError:
            name = ""

        return "<%s, name=%s, val=%r>" % (self.__class__.__name__, name, val)

    def __bool__(self):
        v = evalParam(self)
        assert v._isFullVld()
        return bool(v.val)

    def __int__(self):
        v = evalParam(self)
        assert v._isFullVld()
        return int(v.val)


def evalParam(p):
    """
    Get value of parameter
    """
    while isinstance(p, Param):
        p = p.get()

    if isinstance(p, RtlSignalBase):
        return p.staticEval()
        # use rather param inheritance instead of param as param value
    return toHVal(p)


def getParam(p):
    """
    get param value or value
    getParam gets only item which is specified as param value,
    but evalParam also evaluates it if it is not value
    """
    if isinstance(p, Param):
        p = p.get()
        # use rather param inheritance instead of param as param value
    return p



from hwt.synthesizer.exceptions import IntfLvlConfErr
from hwt.synthesizer.interfaceLevel.interfaceUtils.utils import walkParams
from hwt.synthesizer.interfaceLevel.mainBases import UnitBase
from hwt.synthesizer.interfaceLevel.propDeclrCollector import PropDeclrCollector
from hwt.synthesizer.interfaceLevel.unitImplHelpers import UnitImplHelpers
from hwt.synthesizer.rtlLevel.netlist import RtlNetlist


class Unit(UnitBase, PropDeclrCollector, UnitImplHelpers):
    """
    Container of the netlist with interfaces
    and internal hierarchical structure

    :cvar _serializeDecision: function to decide if Hdl object derived from
        this unit should be serialized or not, if None all is always serialized
    :ivar _interfaces: all interfaces
    :ivar _units: all units defined on this obj
    :ivar _params: all params defined on this obj
    :ivar _parent: parent object (Unit instance)
    :ivar _lazyLoaded: container of rtl object which were lazy loaded
        in implementation phase (this object has to be returned
        from _toRtl of parent before it it's own objects)
    :ivar _targetPlatform: metainformations about target platform
    """

    _serializeDecision = None
    _PROTECTED_NAMES = set(["_PROTECTED_NAMES", "_interfaces",
                            "_units", "_params", "_parent",
                            "_lazyLoaded", "_cntx",
                            "_externInterf", "_targetPlatform"])

    def __init__(self):
        self._parent = None
        self._lazyLoaded = []
        self._cntx = RtlNetlist(self)

        self._loadConfig()

    def _toRtl(self, targetPlatform):
        """
        synthesize all subunits, make connections between them,
        build entity and component for this unit
        """
        assert not self._wasSynthetised()

        self._targetPlatform = targetPlatform
        if not hasattr(self, "_name"):
            self._name = self._getDefaultName()
        self._cntx.params = self._buildParams()
        self._externInterf = []

        # prepare subunits
        for u in self._units:
            yield from u._toRtl(targetPlatform)

        for u in self._units:
            subUnitName = u._name
            u._signalsForMyEntity(self._cntx, "sig_" + subUnitName)

        # prepare signals for interfaces
        for i in self._interfaces:
            signals = i._signalsForInterface(self._cntx)
            if i._isExtern:
                self._externInterf.extend(signals)

        self._loadMyImplementations()
        yield from self._lazyLoaded

        if not self._externInterf:
            raise IntfLvlConfErr(
                "Can not find any external interface for unit %s"
                "- unit without interfaces are not allowed"
                % self._name)

        yield from self._synthetiseContext(self._externInterf)
        self._checkArchCompInstances()
        for intf in self._interfaces:
            intf._setDirLock(True)

    def _wasSynthetised(self):
        return self._cntx.synthesised

    def _synthetiseContext(self, externInterf):
        # synthesize signal level context
        s = self._cntx.synthesize(self._name, externInterf)
        self._entity = s[0]
        self._entity.__doc__ = self.__doc__
        self._entity.origin = self

        self._architecture = s[1]

        for intf in self._interfaces:
            if intf._isExtern:
                intf._resolveDirections()
                # reverse because other components
                # looks at this one from outside
                intf._reverseDirection()

        # connect results of synthesized context to interfaces of this unit
        self._boundInterfacesToEntity(self._interfaces)
        yield from s

        # after synthesis clean up interface so unit can be used elsewhere
        self._cleanAsSubunit()

    def _loadInterface(self, i, isExtern):
        i._loadDeclarations()
        i._setAsExtern(isExtern)

    def _loadDeclarations(self):
        """
        Load all declarations from _decl() method, recursively
        for all interfaces/units.
        """
        if not hasattr(self, "_interfaces"):
            self._interfaces = []
        if not hasattr(self, "_units"):
            self._units = []
        self._setAttrListener = self._declrCollector
        self._declr()
        self._setAttrListener = None
        for i in self._interfaces:
            self._loadInterface(i, True)

        # if I am a unit load subunits
        for u in self._units:
            u._loadDeclarations()
        for p in self._params:
            p.setReadOnly()

    def _registerIntfInImpl(self, iName, intf):
        """
        Register interface in implementation phase
        """
        self._registerInterface(iName, intf)
        self._loadInterface(intf, False)
        intf._signalsForInterface(self._cntx)

    def _buildParams(self):
        # construct params for entity (generics)
        globalNames = {}

        def addP(n, p):
            p.name = n
            if n in globalNames:
                raise IntfLvlConfErr(
                    "Redefinition of generic '%s' while synthesis"
                    " old:%r, new:%r"
                    % (n, globalNames[n], p))
            globalNames[n] = p

        def nameForNestedParam(p):
            n = ""
            node = p
            while node is not self:
                if n == "":
                    n = node._name
                else:
                    n = node._name + "_" + n
                node = node._parent

            return n

        discoveredParams = set()
        for p in self._params:
            discoveredParams.add(p)
            addP(p.name, p)

        for intf in self._interfaces:
            for p in walkParams(intf, discoveredParams):
                n = nameForNestedParam(p)
                addP(n, p)

        return globalNames

    def _getDefaultName(self):
        return self.__class__.__name__

    def _checkArchCompInstances(self):
        cInstances = len(self._architecture.componentInstances)
        units = len(self._units)
        if cInstances != units:
            inRtl = set(
                map(lambda x: x.name, self._architecture.componentInstances))
            inIntf = set(map(lambda x: x._name + "_inst", self._units))
            if cInstances > units:
                raise IntfLvlConfErr(
                    "_toRtl unit(s) %s were found in rtl but were"
                    " not registered at %s"
                    % (str(inRtl - inIntf), self._name))
            elif cInstances < units:
                raise IntfLvlConfErr("_toRtl of %s: unit(s) %s were lost"
                                     % (self._name, str(inIntf - inRtl)))



import xml.dom.minidom

from hwt.pyUtils.arrayQuery import where
import xml.etree.ElementTree as etree


ns = {"spirit": "http://www.spiritconsortium.org/XMLSchema/SPIRIT/1685-2009",
      "xilinx": "http://www.xilinx.com",
      "xsi": "http://www.w3.org/2001/XMLSchema-instance"}
spi_ns_prefix = "{" + ns["spirit"] + "}"
xi_ns_prefix = "{" + ns["xilinx"] + "}"


def findS(elm, name):
    return elm.find("spirit:" + name, ns)


def prettify(elm):
    # or xml.dom.minidom.parseString(xml_string)
    s = xml.dom.minidom.parseString(etree.tostring(elm))
    pretty_xml_as_string = s.toprettyxml()
    return pretty_xml_as_string


def mkSpiElm(elemName):
    e = etree.Element(spi_ns_prefix + elemName)
    return e


def mkXiElm(elemName):
    e = etree.Element(xi_ns_prefix + elemName)
    return e


def appendSpiElem(root, elemName):
    e = mkSpiElm(elemName)
    root.append(e)
    return e


def appendXiElem(root, elemName):
    e = mkXiElm(elemName)
    root.append(e)
    return e


def appendStrElements(root, obj, reqPropNames=[], optPropNames=[]):
    for p in reqPropNames:
        e = appendSpiElem(root, p)
        e.text = getattr(obj, p)
    for p in optPropNames:
        if hasattr(obj, p):
            e = appendSpiElem(root, p)
            e.text = getattr(obj, p)


def whereEndsWithExt(files, extension):
    return where(files, lambda x: x.lower().endswith(extension))


def whereEndsWithExts(files, extensions):
    for f in files:
        for extension in extensions:
            if f.lower().endswith(extension.lower()):
                yield f


def appendSpiAtribs(obj, elm, prefix='', reqPropNames=[], optPropNames=[]):
    for p in reqPropNames:
        elm.attrib[prefix + p] = getattr(obj, p)
    for p in optPropNames:
        if hasattr(obj, p):
            elm.attrib[prefix + p] = getattr(obj, p)


def appendSpiArray(root, arrName, arr):
    if arr:
        elmArr = appendSpiElem(root, arrName)
        for o in arr:
            elmArr.append(o.asElem())



from hwt.hdl.types.integerVal import IntegerVal
from hwt.synthesizer.param import evalParam
from hwt.serializer.simModel.types import SimModelSerializer_types


class HwtSerializer_types(SimModelSerializer_types):
    """
    part of SimModelSerializer responsible for type serialization
    """

    @classmethod
    def HdlType_bits(cls, typ, ctx, declaration=False):
        assert not declaration
        w = typ.width
        if isinstance(w, int):
            pass
        else:
            w = evalParam(w)
            assert isinstance(w, IntegerVal)
            assert w._isFullVld()
            w = w.val

        return "Bits(%d, signed=%r, forceVector=%r, negated=%r)" % (
            w, typ.signed, typ.forceVector, typ.negated)



from hwt.hdl.operatorDefs import AllOps
from hwt.hdl.types.defs import BIT, SLICE
from hwt.serializer.exceptions import UnsupportedEventOpErr


class SystemCSerializer_ops():
    opPrecedence = {
        AllOps.NOT: 3,
        AllOps.NEG: 5,
        AllOps.RISING_EDGE: 0,
        AllOps.DIV: 6,
        AllOps.ADD: 7,
        AllOps.SUB: 7,
        AllOps.MUL: 3,
        AllOps.EQ: 10,
        AllOps.NEQ: 10,
        AllOps.AND: 11,
        AllOps.XOR: 12,
        AllOps.OR: 13,
        AllOps.DOWNTO: 2,
        AllOps.TO: 2,
        AllOps.GT: 9,
        AllOps.LT: 9,
        AllOps.GE: 9,
        AllOps.LE: 9,
        AllOps.CONCAT: 5,
        AllOps.INDEX: 1,
        AllOps.TERNARY: 16,
        AllOps.CALL: 2,
        # AllOps.SHIFTL:8,
        # AllOps.SHIFTR:8,
    }

    _unaryOps = {
        AllOps.NOT: "~%s",
        AllOps.BitsToInt: "%s",
    }

    _binOps = {
        AllOps.AND: '%s & %s',
        AllOps.OR: '%s | %s',
        AllOps.XOR: '%s ^ %s',
        AllOps.CONCAT: '%s & %s',
        AllOps.DIV: '%s / %s',
        AllOps.EQ: '%s == %s',
        AllOps.NEQ: '%s != %s',
        AllOps.GT: '%s > %s',
        AllOps.LT: '%s < %s',
        AllOps.GE: '%s >= %s',
        AllOps.LE: '%s <= %s',
        AllOps.SUB: '%s - %s',
        AllOps.MUL: '%s * %s',
        AllOps.ADD: '%s + %s',
    }
    #  AllOps.DOWNTO:  '%s:%s',
    #  AllOps.TO: '%s:%s'

    @classmethod
    def Operator(cls, op, ctx):
        ops = op.operands
        o = op.operator

        op_str = cls._unaryOps.get(o, None)
        if op_str is not None:
            return op_str % (cls._operand(ops[0], o, ctx))

        op_str = cls._binOps.get(o, None)
        if op_str is not None:
            return op_str % (cls._operand(ops[0], o, ctx),
                             cls._operand(ops[1], o, ctx))

        if o == AllOps.INDEX:
            assert len(ops) == 2
            o0, o1 = ops
            o0_str = cls.asHdl(o0, ctx)
            if ops[1]._dtype == SLICE:
                return "%s.range(%s, %s)" % (o0_str,
                                             cls._operand(o1.val[0], o, ctx),
                                             cls._operand(o1.val[1], o, ctx))
            else:
                return "%s[%s]" % (o0_str, cls._operand(o1, o, ctx))

        elif o == AllOps.TERNARY:
            zero, one = BIT.fromPy(0), BIT.fromPy(1)
            if ops[1] == one and ops[2] == zero:
                # ignore redundant x ? 1 : 0
                return cls.condAsHdl([ops[0]], True, ctx)
            else:
                return "%s ? %s : %s" % (cls.condAsHdl([ops[0]], True, ctx),
                                         cls._operand(ops[1], o, ctx),
                                         cls._operand(ops[2], o, ctx))
        elif o == AllOps.RISING_EDGE or o == AllOps.FALLIGN_EDGE:
            if ctx.isSensitivityList:
                if o == AllOps.RISING_EDGE:
                    _o = ".pos()"
                else:
                    _o = ".neg()"
                return cls._operand(ops[0], o, ctx) + _o
            else:
                raise UnsupportedEventOpErr()
        elif o in [AllOps.BitsAsSigned, AllOps.BitsAsUnsigned,
                   AllOps.BitsAsVec, AllOps.IntToBits]:
            assert len(ops) == 1
            return "static_cast<%s>(%s)" % (cls.HdlType(op.result._dtype, ctx),
                                            cls._operand(ops[0], o, ctx))
        elif o == AllOps.POW:
            assert len(ops) == 2
            raise NotImplementedError()
            # return _bin('**')
        elif o == AllOps.CALL:
            return "%s(%s)" % (cls.FunctionContainer(ops[0]),
                               ", ".join(map(lambda op: cls._operand(op, o, ctx), ops[1:])))
        else:
            raise NotImplementedError(
                "Do not know how to convert %s to vhdl" % (o))



import math
from operator import and_, or_, xor, add

from hwt.hdl.operatorDefs import concatFn
from hwt.hdl.typeShortcuts import hInt, vec
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.typeCast import toHVal
from hwt.pyUtils.arrayQuery import arr_any, flatten
from hwt.synthesizer.andReducedContainer import AndReducedContainer
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.rtlLevel.mainBases import RtlSignalBase
from hwt.synthesizer.rtlLevel.signalUtils.walkers import \
    discoverEventDependency
from hwt.synthesizer.vectorUtils import fitTo


def _intfToSig(obj):
    if isinstance(obj, InterfaceBase):
        return obj._sig
    else:
        return obj


class StmCntx(list):
    "Base class of statement contexts"


class If(StmCntx):
    """
    If statement generator

    :ivar nowIsEventDependent: flag if current scope of if is event dependent
        (is used to mark statements as event dependent)
    """

    def __init__(self, cond, *statements):
        """
        :param cond: condition in if statement
        :param statements: list of statements which should be active
            if condition is met
        """
        self.cond = _intfToSig(cond)
        assert isinstance(self.cond, RtlSignalBase)

        self.nowIsEventDependent = arr_any(discoverEventDependency(cond),
                                           lambda x: True)
        self.elifConds = []

        c = AndReducedContainer()
        c.add(self.cond)
        self._appendStatements(c, statements)

    def Else(self, *statements):
        ncond = AndReducedContainer()

        for ec in reversed(self.elifConds):
            ncond.add(~ec)
        ncond.add(~self.cond)

        self._appendStatements(ncond, statements)

        # convert self to StmCntx to prevent any other else/elif
        stml = StmCntx()
        stml.extend(self)
        return stml

    def _appendStatements(self, condSet, statements):
        """
        Append statements to this container under conditions specified
        by condSet
        """
        for stm in flatten(statements):
            stm.isEventDependent = (stm.isEventDependent
                                    or self.nowIsEventDependent)
            for c in condSet:
                c.endpoints.append(stm)
            stm.cond.update(condSet)
            self.append(stm)

    def Elif(self, cond, *statements):
        cond = _intfToSig(cond)
        self.nowIsEventDependent = (self.nowIsEventDependent or
                                    arr_any(discoverEventDependency(cond),
                                            lambda x: True))
        thisCond = AndReducedContainer()
        thisCond.add(cond)
        for c in reversed(self.elifConds):
            thisCond.add(~c)

        thisCond.add(~self.cond)
        self._appendStatements(thisCond, statements)
        self.elifConds.append(cond)

        return self


class Switch(StmCntx):
    """
    Switch statement generator
    """

    def __init__(self, switchOn):
        self.switchOn = switchOn
        self.cond = None

    _appendStatements = If._appendStatements

    def Case(self, caseVal, *statements):
        "c-like case of switch statement"
        cond = self.switchOn._eq(caseVal)
        if self.cond is None:
            If.__init__(self, cond, *statements)
        else:
            If.Elif(self, cond, *statements)

        return self

    def addCases(self, tupesValStmnts):
        """
        Add multiple case statements from iterable of tuleles (caseVal, statements)
        """
        s = self
        for val, statements in tupesValStmnts:
            s = s.Case(val, *statements)
        return s

    def Default(self, *statements):
        """c-like default of switch statement
        """
        if self.cond is None:
            # no cases were used
            return statements

        return If.Else(self, *statements)


def SwitchLogic(cases, default=None):
    """
    generate if tree for cases like

    ..code-block:: python
        if cond0:
            statements0
        elif cond1:
            statements1
        else:
            default

    :param case: iterable of tuples (condition, statements)
    :param default: default statements
    """
    if default is not None:
        assigTop = default
    else:
        assigTop = []

    for cond, statements in reversed(cases):
        assigTop = If(cond,
                      statements
                      ).Else(
            assigTop
        )

    return assigTop


def In(sigOrVal, iterable):
    """
    Hdl convertible in operator, check if any of items
    in "iterable" equals "sigOrVal"
    """
    res = None
    for i in iterable:
        i = toHVal(i)
        if res is None:
            res = sigOrVal._eq(i)
        else:
            res = res | sigOrVal._eq(i)

    assert res is not None, "Parameter iterable is empty"
    return res


def StaticForEach(parentUnit, items, bodyFn, name=""):
    """
    Generate for loop for static items

    :param parentUnit: unit where this code should be instantiated
    :param items: items which this "for" itering on
    :param bodyFn: function which fn(item, index) or fn(item)
        returns (statementList, ack).
        It's content is performed in every iteration.
        When ack is high loop will fall to next iteration
    """

    items = list(items)
    itemsCnt = len(items)
    if itemsCnt == 0:
        # if there are no items there is nothing to generate
        return []
    elif itemsCnt == 1:
        # if there is only one item do not generate counter logic generate
        return bodyFn(items[0], 0)
    else:
        # if there is multiple items we have to generate counter logic
        index = parentUnit._reg(name + "for_index",
                                Bits(log2ceil(itemsCnt + 1), signed=False),
                                defVal=0)
        ackSig = parentUnit._sig(name + "for_ack")

        statementLists = []
        for i, (statementList, ack) in [(i, bodyFn(item, i))
                                        for i, item in enumerate(items)]:
            statementLists.append(statementList + [(ackSig(ack)), ])

        If(ackSig,
           If(index._eq(itemsCnt - 1),
              index(0)
              ).Else(
               index(index + 1)
           )
        )

        return Switch(index)\
            .addCases(
            enumerate(statementLists)
        ).Default(
            bodyFn(items[0], 0)[0]
        )


class FsmBuilder(StmCntx):
    """
    :ivar stateReg: register with state
    """

    def __init__(self, parent, stateT, stateRegName="st"):
        """
        :param parent: parent unit where fsm should be builded
        :param stateT: enum type of state
        :param stateRegName: name of register where sate is stored
        """
        if isinstance(stateT, HEnum):
            beginVal = stateT.fromPy(stateT._allValues[0])
        else:
            beginVal = 0

        self.stateReg = parent._reg(stateRegName, stateT, beginVal)
        Switch.__init__(self, self.stateReg)

    _appendStatements = Switch._appendStatements

    def Trans(self, stateFrom, *condAndNextState):
        """
        :param stateFrom: apply when FSM is in this state
        :param condAndNextState: tupes (condition, newState),
            last does not to have condition

        :attention: transitions has priority, first has the biggest
        :attention: if stateFrom is None it is evaluated as default
        """
        top = None
        last = True

        for cAndS in reversed(condAndNextState):
            if last is True:
                last = False
                # if this is last trans. it does not have to condition
                try:
                    condition, newvalue = cAndS
                except TypeError:
                    top = c(cAndS, self.stateReg)
                    continue
                top = []

            else:
                condition, newvalue = cAndS

            # building decision tree
            top = If(condition,
                     self.stateReg(newvalue)
                     ).Else(
                top
            )

        s = Switch.Case(self, stateFrom, *top)
        return s

    def Default(self, *condAndNextState):
        d = self.Trans(None, *condAndNextState)
        d.stateReg = self.stateReg
        return d


def _connect(src, dst, exclude, fit):
    if isinstance(src, InterfaceBase):
        if isinstance(dst, InterfaceBase):
            return dst._connectTo(src, exclude=exclude, fit=fit)
        src = src._sig

    assert not exclude, "this intf. is just a signal"
    if src is None:
        src = dst._dtype.fromPy(None)
    else:
        src = toHVal(src)

    if fit:
        src = fitTo(src, dst)

    src = src._auto_cast(dst._dtype)

    return dst(src)


def connect(src, *destinations, exclude=set(), fit=False):
    """
    Connect src (signals/interfaces/values) to all destinations

    :param exclude: interfaces on any level on src or destinations
        which should be excluded from connection process
    :param fit: auto fit source width to destination width
    """
    assignemnts = []
    for dst in destinations:
        assignemnts.extend(_connect(src, dst, exclude, fit))

    return assignemnts


def _mkOp(fn):
    """
    Function to create variadic operator function

    :param fn: function to perform binary operation
    """
    def op(*operands, key=None) -> RtlSignalBase:
        """
        :param operands: variadic parameter of input uperands
        :param key: optional function applied on every operand
            before processing
        """
        assert operands, operands
        top = None
        if key is not None:
            operands = map(key, operands)

        for s in operands:
            if top is None:
                top = s
            else:
                top = fn(top, s)
        return top

    return op


# variadic operator functions
And = _mkOp(and_)
Add = _mkOp(add)
Or = _mkOp(or_)
Xor = _mkOp(xor)
Concat = _mkOp(concatFn)


def power(base, exp) -> RtlSignalBase:
    return toHVal(base)._pow(exp)


def ror(sig, howMany) -> RtlSignalBase:
    "Rotate right"
    return sig[howMany:]._concat(sig[:howMany])


def rol(sig, howMany) -> RtlSignalBase:
    "Rotate left"
    width = sig._dtype.bit_length()
    return sig[(width - howMany):]._concat(sig[:(width - howMany)])


def sll(sig, howMany) -> RtlSignalBase:
    "Logical shift left"
    width = sig._dtype.bit_length()
    return sig[(width - howMany):]._concat(vec(0, howMany))


def srl(sig, howMany) -> RtlSignalBase:
    "Logical shift right"
    return vec(0, howMany)._concat(sig[:howMany])


def log2ceil(x):
    """
    Returns no of bits required to store x-1
    for example x=8 returns 3
    """

    if not isinstance(x, (int, float)):
        x = int(x)

    if x == 0 or x == 1:
        res = 1
    else:
        res = math.ceil(math.log2(x))

    return hInt(res)


def isPow2(num) -> bool:
    if not isinstance(num, int):
        num = int(num)
    return num != 0 and ((num & (num - 1)) == 0)


def binToGray(sigOrVal) -> RtlSignalBase:
    width = sigOrVal._dtype.bit_length()
    return Concat(sigOrVal[width - 1],
                  sigOrVal[width - 1:0] ^ sigOrVal[width:1])


def sizeof(_type) -> int:
    "get size of type in bytes"
    s = _type.bit_length()
    return math.ceil(s / 8)


# shortcuts
c = connect



from itertools import chain

from hwt.hdl.assignment import Assignment
from hwt.hdl.statements import WaitStm, IfContainer, \
    SwitchContainer


def getMaxStmIdForStm(stm):
    maxId = 0
    if isinstance(stm, Assignment):
        return stm._instId
    elif isinstance(stm, IfContainer):
        for _stm in chain(stm.ifTrue, *map(lambda _elif: _elif[1], stm.elIfs),
                          stm.ifFalse):
            maxId = max(maxId, getMaxStmIdForStm(_stm))
        return maxId
    elif isinstance(stm, SwitchContainer):
        for _stm in chain(*map(lambda _case: _case[1], stm.cases)):
            maxId = max(maxId, getMaxStmIdForStm(_stm))
        return maxId
    elif isinstance(stm, WaitStm):
        return maxId
    else:
        raise NotImplementedError(stm)


def maxStmId(proc):
    """
    get max statement id,
    used for sorting of processes in architecture
    """
    maxId = 0
    for stm in proc.statements:
        maxId = max(maxId, getMaxStmIdForStm(stm))
    return maxId



from itertools import zip_longest
from math import ceil, floor, inf
import re
from typing import Union, Generator

from hwt.bitmask import mask, selectBitRange, setBitRange
from hwt.hdl.frameTmplUtils import TransTmplWordIterator, \
    ChoicesOfFrameParts
from hwt.hdl.transPart import TransPart
from hwt.hdl.types.array import HArray
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.struct import HStruct
from hwt.pyUtils.arrayQuery import flatten
from hwt.simulator.types.simBits import simBitsT


class FrameTmpl(object):
    """
    Frame template is metainfomation about data structure, it's the template
    of transaction for specific interface

    Usuall flow of frame generatig consists of these steps:

    1. Describe format of data by HDL type (HStruct, HUnion ...)

    2. Convert it to TransTmpl to resolve addresses of each field in structure

    3. Split parts of TransTmpl into words

    4. Use parts in words to assembly frames

    :ivar _fieldToTPart: dictionary {HStructField: TransPart} to resolve
        this association, None by default, builded when packData
        is called and is not builded
    :note: others ivars described in __init__
    """
    __RE_RM_ARRAY_DOTS = re.compile("(\.\[)")

    def __init__(self, origin, wordWidth, startBitAddr, endBitAddr,
                 transParts):
        """
        :param origin: instance of HType (usually HStruct)
            from which this FrameTmpl was generated from
        :param wordWidth: width of word on interface
            where this template should be used
        :param startBitAddr: bit offset where this frame starts
        :param endBitAddr: bit offset where this frame ends
            (bit index of first bit behind this frame)
        :param transParts: instances of TransPart which are parts of this frame
        """
        self.origin = origin
        self.wordWidth = wordWidth
        assert startBitAddr <= endBitAddr
        self.startBitAddr = startBitAddr
        self.endBitAddr = endBitAddr
        self.parts = transParts

        self._fieldToTPart = None

        for p in self.parts:
            p.parent = self
            assert p.startOfPart >= startBitAddr, (p, startBitAddr)
            assert p.endOfPart <= endBitAddr, (p, endBitAddr)

    @staticmethod
    def framesFromTransTmpl(transaction: 'TransTmpl',
                            wordWidth: int,
                            maxFrameLen: Union[int, float]=inf,
                            maxPaddingWords: Union[int, float]=inf,
                            trimPaddingWordsOnStart: bool=False,
                            trimPaddingWordsOnEnd: bool=False) -> Generator[
                                'FrameTmpl', None, None]:
        """
        Convert transaction template into FrameTmpls

        :param transaction: transaction template used which are FrameTmpls
            created from
        :param wordWidth: width of data signal in target interface
            where frames will be used
        :param maxFrameLen: maximum length of frame in bits,
            if exceeded another frame will be created
        :param maxPaddingWords: maximum of continual padding words in frame,
            if exceed frame is split and words are cut of
        :attention: if maxPaddingWords<inf trimPaddingWordsOnEnd
            or trimPaddingWordsOnStart has to be True
            to decide where padding should be trimmed
        :param trimPaddingWordsOnStart: trim padding from start of frame
            at word granularity
        :param trimPaddingWordsOnEnd: trim padding from end of frame
            at word granularity
        """
        isFirstInFrame = True
        partsPending = False

        startOfThisFrame = 0
        assert maxFrameLen > 0
        assert maxPaddingWords >= 0
        if maxPaddingWords < inf:
            assert trimPaddingWordsOnStart or trimPaddingWordsOnEnd, \
                "Padding has to be cut off somewhere"

        it = TransTmplWordIterator(wordWidth)
        lastWordI = 0
        endOfThisFrame = maxFrameLen
        parts = []
        for wordI, word in it.groupByWordIndex(transaction, 0):
            if wordI * wordWidth >= endOfThisFrame:
                # now in first+ word behind the frame
                # cut off padding at end of frame
                paddingWords = wordI - lastWordI
                if trimPaddingWordsOnEnd and paddingWords > maxPaddingWords:
                    # cut off padding and align end of frame to word
                    _endOfThisFrame = (lastWordI + 1) * wordWidth
                else:
                    _endOfThisFrame = wordI * wordWidth

                yield FrameTmpl(transaction,
                                wordWidth,
                                startOfThisFrame,
                                _endOfThisFrame,
                                parts)

                # prepare for start of new frame
                parts = []
                isFirstInFrame = True
                partsPending = False
                # start on new word
                startOfThisFrame = _endOfThisFrame
                endOfThisFrame = startOfThisFrame + maxFrameLen
                lastWordI = wordI

            # check if padding at potential end of frame can be cut off
            if (not isFirstInFrame
                    and trimPaddingWordsOnEnd
                    and wordI - lastWordI > 1):
                # there is too much continual padding,
                # cut it out and start new frame
                _endOfThisFrame = (lastWordI + 1) * wordWidth
                yield FrameTmpl(transaction,
                                wordWidth,
                                startOfThisFrame,
                                _endOfThisFrame,
                                parts)

                # prepare for start of new frame
                parts = []
                isFirstInFrame = True
                partsPending = False
                # start on new word
                startOfThisFrame = _endOfThisFrame
                endOfThisFrame = startOfThisFrame + maxFrameLen
                lastWordI = wordI - 1

            if isFirstInFrame:
                partsPending = True
                isFirstInFrame = False
                # cut off padding at start of frame
                paddingWords = wordI - lastWordI
                if trimPaddingWordsOnStart and paddingWords > maxPaddingWords:
                    startOfThisFrame += paddingWords * wordWidth

                endOfThisFrame = startOfThisFrame + maxFrameLen

            # resolve end of this part
            parts.extend(word)
            lastWordI = wordI

        # reminder in "parts" after last iteration
        endOfThisFrame = transaction.bitAddrEnd
        withPadding = not (trimPaddingWordsOnEnd or trimPaddingWordsOnStart)
        if partsPending or (withPadding
                            and endOfThisFrame != startOfThisFrame):
            # cut off padding at end of frame
            endOfLastWord = (lastWordI + 1) * wordWidth
            if endOfThisFrame < endOfLastWord:
                endOfThisFrame = endOfLastWord
            else:
                paddingWords = it.fullWordCnt(endOfLastWord, endOfThisFrame)
                if trimPaddingWordsOnEnd and paddingWords > maxPaddingWords:
                    endOfThisFrame -= paddingWords * wordWidth
                    # align end of frame to word
            endOfThisFrame = min(startOfThisFrame +
                                 maxFrameLen, endOfThisFrame)

            yield FrameTmpl(transaction,
                            wordWidth,
                            startOfThisFrame,
                            endOfThisFrame,
                            parts)
            parts = []
            startOfThisFrame = endOfThisFrame

        # final padding on the end
        while withPadding and startOfThisFrame < transaction.bitAddrEnd:
            endOfThisFrame = min(startOfThisFrame +
                                 maxFrameLen, transaction.bitAddrEnd)

            yield FrameTmpl(transaction,
                            wordWidth,
                            startOfThisFrame,
                            endOfThisFrame,
                            [])

            startOfThisFrame = endOfThisFrame

    def _wordIndx(self, addr: int):
        """
        convert bit address to index of word where this address is
        """
        return floor(addr / self.wordWidth)

    def getWordCnt(self):
        """
        Get count of words in this frame
        """
        return ceil((self.endBitAddr - self.startBitAddr) / self.wordWidth)

    def walkWords(self, showPadding: bool=False):
        """
        Walk enumerated words in this frame

        :attention: not all indexes has to be present, only words
            with items will be generated when not showPadding
        :param showPadding: padding TransParts are also present
        :return: generator of tuples (wordIndex, list of TransParts
            in this word)
        """
        wIndex = 0
        lastEnd = self.startBitAddr
        parts = []
        for p in self.parts:
            end = p.startOfPart
            if showPadding and end != lastEnd:
                # insert padding
                while end != lastEnd:
                    assert end >= lastEnd, (end, lastEnd)
                    endOfWord = ceil(
                        (lastEnd + 1) / self.wordWidth) * self.wordWidth
                    endOfPadding = min(endOfWord, end)
                    _p = TransPart(self, None, lastEnd, endOfPadding, 0)
                    parts.append(_p)

                    if endOfPadding >= endOfWord:
                        yield (wIndex, parts)
                        wIndex += 1
                        parts = []

                    lastEnd = endOfPadding

            if self._wordIndx(lastEnd) != self._wordIndx(p.startOfPart):
                yield (wIndex, parts)

                wIndex += 1
                parts = []
                lastEnd = p.endOfPart

            parts.append(p)
            lastEnd = p.endOfPart
            if lastEnd % self.wordWidth == 0:
                yield (wIndex, parts)

                wIndex += 1
                parts = []

        if showPadding and (parts
                            or lastEnd != self.endBitAddr
                            or lastEnd % self.wordWidth != 0):
            # align end to end of last word
            end = ceil(self.endBitAddr / self.wordWidth) * self.wordWidth
            while end != lastEnd:
                assert end >= lastEnd, (end, lastEnd)
                endOfWord = ((lastEnd // self.wordWidth) + 1) * self.wordWidth
                endOfPadding = min(endOfWord, end)
                _p = TransPart(self, None, lastEnd, endOfPadding, 0)
                _p.parent = self
                parts.append(_p)

                if endOfPadding >= endOfWord:
                    yield (wIndex, parts)
                    wIndex += 1
                    parts = []

                lastEnd = endOfPadding

            if parts:
                # in the case end of frame is not aligned to end of word
                yield (wIndex, parts)

    @staticmethod
    def fieldToDataDict(dtype, data, res):
        """
        Construct dictionary {StructField:value} for faster lookup of values
        for fields
        """
        # assert data is None or isinstance(data, dict)
        for f in dtype.fields:
            try:
                fVal = data[f.name]
            except KeyError:
                fVal = None

            if isinstance(f.dtype, Bits):
                if fVal is not None:
                    assert isinstance(fVal, int)
                    res[f] = fVal
            elif isinstance(f.dtype, HStruct):
                if fVal:
                    FrameTmpl.fieldToDataDict(f.dtype, fVal, res)
            elif isinstance(f.dtype, HArray):
                if fVal:
                    # assert isinstance(fVal, class_or_tuple)
                    res[f] = fVal

        return res

    def packData(self, data):
        """
        Pack data into list of BitsVal of specified dataWidth

        :param data: dict of values for struct fields {fieldName: value}

        :return: list of BitsVal which are representing values of words
        """
        typeOfWord = simBitsT(self.wordWidth, None)
        fieldToVal = self._fieldToTPart
        if fieldToVal is None:
            fieldToVal = self._fieldToTPart = self.fieldToDataDict(
                self.origin.dtype,
                data,
                {})

        for _, transParts in self.walkWords(showPadding=True):
            actualVldMask = 0
            actualVal = 0
            for tPart in transParts:
                high, low = tPart.getBusWordBitRange()
                fhigh, flow = tPart.getFieldBitRange()
                if not tPart.isPadding:
                    val = fieldToVal.get(tPart.tmpl.origin, None)
                else:
                    val = None

                if val is None:
                    newBits = 0
                    vld = 0
                else:
                    newBits = selectBitRange(val, flow, fhigh - flow)
                    vld = mask(high - low) << low

                actualVal = setBitRange(actualVal, low, high - low, newBits)
                actualVldMask = setBitRange(actualVal, low, high - low, vld)

            yield typeOfWord.getValueCls()(actualVal, typeOfWord,
                                           actualVldMask, -1)

    def __repr__getName(self, transPart, fieldWidth):
        if transPart.isPadding:
            return "X" * fieldWidth
        else:
            names = []
            tp = transPart.tmpl
            while tp is not None:
                try:
                    isArrayElm = isinstance(tp.parent.dtype, HArray)
                except AttributeError:
                    isArrayElm = False

                if isArrayElm:
                    arr = transPart.tmpl.parent
                    arrS = arr.bitAddr
                    itemW = (arr.bitAddrEnd - arrS) // arr.itemCnt
                    s = transPart.startOfPart
                    indx = (s - arrS) // itemW
                    names.append("[%d]" % indx)
                else:
                    o = tp.origin
                    if o is None:
                        break
                    if o.name is not None:
                        names.append(o.name)
                tp = tp.parent

            # [HOTFIX] rm dots when indexing on array
            return self.__RE_RM_ARRAY_DOTS.sub("[", ".".join(reversed(names)))

    def __repr__word(self, index, width, padding, transParts):
        buff = ["{0: <{padding}}|".format(index, padding=padding)]
        DW = self.wordWidth
        partsWithChoice = []

        endAlignment = transParts[-1].endOfPart % DW
        if endAlignment:
            # -1 for ending |
            percentOfWidth = (DW - endAlignment) / DW
            # -1 for ending |
            fieldWidth = max(0, int(percentOfWidth * width) - 1)
            assert fieldWidth >= 0
            s = '%s|' % ("^" * (fieldWidth))
            buff.append(s)

        for tp in reversed(transParts):
            percentOfWidth = tp.bit_length() / DW
            # -1 for ending |
            fieldWidth = max(0, int(percentOfWidth * width) - 1)
            assert fieldWidth >= 0

            # percentOffset = (tp.inFrameBitAddr % DW) / DW
            # offset = int(percentOffset * width)
            if isinstance(tp, ChoicesOfFrameParts):
                name = "<union>"
                partsWithChoice.append(tp)
            else:
                name = self.__repr__getName(tp, fieldWidth)
            buff.append('{0: ^{fieldWidth}}|'.format(
                name, fieldWidth=fieldWidth))
        return ("".join(buff), partsWithChoice)

    def __repr__(self, scale=1):
        buff = []
        s = "<%s start:%d, end:%d" % (
            self.__class__.__name__, self.startBitAddr, self.endBitAddr)
        buff.append(s)

        padding = 5
        DW = self.wordWidth
        width = int(DW * scale)

        buff.append(
            '{0: <{padding}}{1: <{halfLineWidth}}{2: >{halfLineWidth}}'.format(
                "", DW - 1, 0, padding=padding, halfLineWidth=width // 2))
        line = '{0: <{padding}}{1:-<{lineWidth}}'.format(
            "", "", padding=padding, lineWidth=width + 1)
        buff.append(line)

        for w, transParts in self.walkWords(showPadding=True):
            wStr, partsWithChoice = self.__repr__word(
                w, width, padding, transParts)
            buff.append(wStr)
            while partsWithChoice:
                for parts in zip_longest(*partsWithChoice):
                    parts = list(flatten(parts, level=1))
                    wStr, _partsWithChoice = self.__repr__word(
                        w, width, padding, parts)
                    buff.append(wStr)
                partsWithChoice = _partsWithChoice

        buff.append(line)
        buff.append(">")
        return "\n".join(buff)



from copy import copy
from jinja2.environment import Environment
from jinja2.loaders import PackageLoader

from hwt.hdl.architecture import Architecture
from hwt.hdl.assignment import Assignment
from hwt.hdl.constants import SENSITIVITY
from hwt.hdl.operator import Operator
from hwt.hdl.operatorDefs import AllOps, sensitivityByOp
from hwt.hdl.statements import IfContainer
from hwt.hdl.types.bits import Bits
from hwt.hdl.types.enum import HEnum
from hwt.hdl.types.enumVal import HEnumVal
from hwt.hdl.types.typeCast import toHVal
from hwt.serializer.exceptions import SerializerException
from hwt.serializer.generic.serializer import GenericSerializer
from hwt.serializer.generic.constCache import ConstCache
from hwt.serializer.generic.indent import getIndent
from hwt.serializer.generic.nameScope import LangueKeyword
from hwt.serializer.simModel.keywords import SIMMODEL_KEYWORDS
from hwt.serializer.simModel.ops import SimModelSerializer_ops
from hwt.serializer.simModel.types import SimModelSerializer_types
from hwt.serializer.simModel.value import SimModelSerializer_value
from hwt.serializer.utils import maxStmId
from hwt.synthesizer.param import evalParam


env = Environment(loader=PackageLoader('hwt', 'serializer/simModel/templates'))
unitTmpl = env.get_template('modelCls.py.template')
processTmpl = env.get_template('process.py.template')
ifTmpl = env.get_template("if.py.template")


class SimModelSerializer(SimModelSerializer_value, SimModelSerializer_ops,
                         SimModelSerializer_types, GenericSerializer):
    """
    Serializer which converts Unit instances to simulator code
    """
    _keywords_dict = {kw: LangueKeyword() for kw in SIMMODEL_KEYWORDS}
    fileExtension = '.py'

    @classmethod
    def serializationDecision(cls, obj, serializedClasses,
                              serializedConfiguredUnits):
        # we need all instances for simulation
        return True

    @classmethod
    def stmAsHdl(cls, obj, ctx, enclosure=None):
        try:
            serFn = getattr(cls, obj.__class__.__name__)
        except AttributeError:
            raise NotImplementedError("Not implemented for %s" % (repr(obj)))
        return serFn(obj, ctx, enclosure=enclosure)

    @classmethod
    def Architecture(cls, arch: Architecture, ctx):
        cls.Entity_prepare(arch.entity, ctx, serialize=False)
        variables = []
        procs = []
        extraTypes = set()
        extraTypes_serialized = []
        arch.variables.sort(key=lambda x: x.name)
        arch.processes.sort(key=lambda x: (x.name, maxStmId(x)))
        arch.componentInstances.sort(key=lambda x: x._name)

        ports = list(
            map(lambda p: (p.name, cls.HdlType(p._dtype, ctx)),
                arch.entity.ports))

        for v in arch.variables:
            t = v._dtype
            # if type requires extra definition
            if isinstance(t, HEnum) and t not in extraTypes:
                extraTypes.add(v._dtype)
                extraTypes_serialized.append(
                    cls.HdlType(t, ctx, declaration=True))

            v.name = ctx.scope.checkedName(v.name, v)
            variables.append(v)

        childCtx = copy(ctx)
        childCtx.constCache = ConstCache(ctx.scope.checkedName)

        def serializeVar(v):
            dv = evalParam(v.defaultVal)
            if isinstance(dv, HEnumVal):
                dv = "%s.%s" % (dv._dtype.name, dv.val)
            else:
                dv = cls.Value(dv, ctx)

            return v.name, cls.HdlType(v._dtype, childCtx), dv

        for p in arch.processes:
            procs.append(cls.HWProcess(p, childCtx))

        constants = []
        for c in sorted(childCtx.constCache._cache.items(), key=lambda x: x[1],
                        reverse=True):
            constants.append((c[1], cls.Value(c[0], ctx)))

        return unitTmpl.render(
            name=arch.getEntityName(),
            constants=constants,
            ports=ports,
            signals=list(map(serializeVar, variables)),
            extraTypes=extraTypes_serialized,
            processes=procs,
            processObjects=arch.processes,
            processesNames=map(lambda p: p.name, arch.processes),
            componentInstances=arch.componentInstances,
            isOp=lambda x: isinstance(x, Operator),
            sensitivityByOp=sensitivityByOp,
            serialize_io=cls.sensitivityListItem,
        )

    @classmethod
    def Assignment(cls, a, ctx, enclosure=None):
        dst = a.dst
        indentStr = getIndent(ctx.indent)
        ev = a.isEventDependent

        srcStr = "%s" % cls.Value(a.src, ctx)
        if a.indexes is not None:
            return "%sio.%s.add((%s, (%s,), %s))" % (
                indentStr, dst.name, srcStr,
                ", ".join(map(lambda x: cls.asHdl(x, ctx),
                              a.indexes)),
                ev)
        else:
            if not (dst._dtype == a.src._dtype):
                srcT = a.src._dtype
                dstT = dst._dtype
                if (isinstance(srcT, Bits) and
                        isinstance(dstT, Bits) and
                        srcT.bit_length() == dstT.bit_length() == 1):

                    if srcT.forceVector != dstT.forceVector:
                        _0 = cls.Value(toHVal(0), ctx)
                        if srcT.forceVector:
                            return "%sio.%s.add(((%s)._getitem__val(%s), %s))"\
                                % (indentStr, dst.name, srcStr, _0, ev)
                        else:
                            return "%sio.%s.add((%s, (%s,), %s))" % (
                                indentStr, dst.name, srcStr, _0, ev)

                raise SerializerException(
                    ("%s <= %s  is not valid assignment\n"
                     " because types are different (%r; %r) ") %
                    (cls.asHdl(dst, ctx), srcStr,
                     dst._dtype, a.src._dtype))
            else:
                return "%sio.%s.add((%s, %s))" % (
                    indentStr, dst.name, srcStr, ev)

    @classmethod
    def comment(cls, comentStr):
        return "#" + comentStr.replace("\n", "\n#")

    @classmethod
    def IfContainer(cls, ifc, ctx, enclosure=None):
        cond = cls.condAsHdl(ifc.cond, ctx)
        ifTrue = ifc.ifTrue
        ifFalse = ifc.ifFalse

        if ifc.elIfs:
            # if has elifs rewind this to tree
            ifFalse = []
            topIf = IfContainer(ifc.cond, ifc.ifTrue, ifFalse)
            for c, stms in ifc.elIfs:
                _ifFalse = []
                lastIf = IfContainer(c, stms, _ifFalse)
                ifFalse.append(lastIf)
                ifFalse = _ifFalse

            lastIf.ifFalse = ifc.ifFalse

            return cls.IfContainer(topIf, ctx, enclosure=enclosure)
        else:
            childCtx = ctx.withIndent()
            if enclosure is None:
                _enclosure = getIndent(childCtx.indent) + "pass"
            else:
                _enclosure = "\n".join(
                    [cls.stmAsHdl(e, childCtx) for e in enclosure])

            return ifTmpl.render(
                indent=getIndent(ctx.indent),
                indentNum=ctx.indent,
                cond=cond,
                enclosure=_enclosure,
                ifTrue=tuple(map(lambda obj: cls.stmAsHdl(obj, childCtx,
                                                          enclosure=enclosure),
                                 ifTrue)),
                ifFalse=tuple(map(lambda obj: cls.stmAsHdl(obj, childCtx,
                                                           enclosure=enclosure),
                                  ifFalse)))

    @classmethod
    def SwitchContainer(cls, sw, ctx, enclosure=None):
        switchOn = sw.switchOn

        def mkCond(c):
            return {Operator(AllOps.EQ,
                             [switchOn, c])}
        elIfs = []

        for key, statements in sw.cases:
            elIfs.append((mkCond(key), statements))
        ifFalse = sw.default

        topCond = mkCond(sw.cases[0][0])
        topIf = IfContainer(topCond,
                            sw.cases[0][1],
                            ifFalse,
                            elIfs)

        return cls.IfContainer(topIf, ctx, enclosure=enclosure)

    @classmethod
    def sensitivityListItem(cls, item):
        if isinstance(item, Operator):
            op = item.operator
            if op == AllOps.RISING_EDGE:
                sens = SENSITIVITY.RISING
            elif op == AllOps.FALLIGN_EDGE:
                sens = SENSITIVITY.FALLING
            else:
                raise TypeError("This is not an event sensitivity", op)

            return "(%s, %s)" % (str(sens), item.operands[0].name)
        else:
            return item.name

    @classmethod
    def HWProcess(cls, proc, ctx):
        body = proc.statements
        proc.name = ctx.scope.checkedName(proc.name, proc)
        sensitivityList = sorted(
            map(cls.sensitivityListItem, proc.sensitivityList))

        childCtx = ctx.withIndent(2)
        if len(body) == 1:
            _body = cls.stmAsHdl(body[0], childCtx)
        else:
            for i, stm in enumerate(body):
                if not isinstance(stm, Assignment):
                    break
            # first statement is taken as default
            enclosure = body[:i]
            _body = "\n".join([cls.stmAsHdl(stm, childCtx, enclosure=enclosure)
                               for stm in body[i:]])

        return processTmpl.render(
            name=proc.name,
            sensitivityList=sensitivityList,
            stmLines=[_body]
        )





class MapExpr():
    """
    Map expression, is used in component instance mapping
    """
    def __init__(self, compSig, value):
        self.compSig = compSig
        self.value = value

    def __repr__(self):
        from hwt.serializer.vhdl.serializer import VhdlSerializer
        return VhdlSerializer.MapExpr(self)




def tryReduceAnd(sig, val):
    """
    Return sig and val reduced by & operator or None
    if it is not possible to statically reduce expression
    """
    m = sig._dtype.all_mask()
    if val._isFullVld():
        v = val.val
        if v == m:
            return sig
        elif v == 0:
            return val


def tryReduceOr(sig, val):
    """
    Return sig and val reduced by | operator or None
    if it is not possible to statically reduce expression
    """
    m = sig._dtype.all_mask()
    if not val.vldMask:
        return val

    if val._isFullVld():
        v = val.val
        if v == m:
            return val
        elif v == 0:
            return sig


def tryReduceXor(sig, val):
    """
    Return sig and val reduced by ^ operator or None
    if it is not possible to statically reduce expression
    """
    m = sig._dtype.all_mask()
    if not val.vldMask:
        return val

    if val._isFullVld():
        v = val.val
        if v == m:
            return ~sig
        elif v == 0:
            return sig



from typing import Tuple

from hwt.hdl.constants import DIRECTION, SENSITIVITY
from hwt.hdl.process import HWProcess
from hwt.hdl.value import Value
from hwt.simulator.utils import valueHasChanged


def sensitivity(proc: HWProcess, *sensitiveTo):
    """
    register sensitivity for process
    """
    for s in sensitiveTo:
        if isinstance(s, tuple):
            sen, s = s
            if sen == SENSITIVITY.ANY:
                s.simSensProcs.add(proc)
            elif sen == SENSITIVITY.RISING:
                s.simRisingSensProcs.add(proc)
            elif sen == SENSITIVITY.FALLING:
                s.simFallingSensProcs.add(proc)
            else:
                raise AssertionError(sen)
        else:
            s.simSensProcs.add(proc)


def simEvalCond(simulator, *conds):
    """
    Evaluate list of values as condition
    """
    _cond = True
    _vld = True
    for v in conds:
        val = bool(v.val)
        fullVld = v.vldMask == 1
        if fullVld:
            if not val:
                return False, True
        else:
            return False, False

        _cond = _cond and val
        _vld = _vld and fullVld

    return _cond, _vld


class SimModel(object):
    """
    Base class for model in simulator
    """


def connectSimPort(simUnit, subSimUnit, srcName, dstName, direction):
    """
    Connect ports of simulation models by name
    """
    if direction == DIRECTION.OUT:
        origPort = getattr(subSimUnit, srcName)
        newPort = getattr(simUnit, dstName)
        setattr(subSimUnit, srcName, newPort)
    else:
        origPort = getattr(subSimUnit, dstName)
        newPort = getattr(simUnit, srcName)
        setattr(subSimUnit, dstName, newPort)

    subSimUnit._cntx.signals.remove(origPort)


def mkUpdater(nextVal: Value, invalidate: bool):
    """
    Create value updater for simulation

    :param nextVal: instance of Value which will be asssiggned to signal
    :param invalidate: flag which tells if value has been compromised
        and if it should be invaidated
    :return: function(value) -> tuple(valueHasChangedFlag, nextVal)
    """

    def updater(currentVal):
        _nextVal = nextVal.clone()
        if invalidate:
            _nextVal.vldMask = 0
        return (valueHasChanged(currentVal, _nextVal), _nextVal)
    return updater


def mkArrayUpdater(nextItemVal: Value, indexes: Tuple[Value],
                   invalidate: bool):
    """
    Create value updater for simulation for value of array type

    :param nextVal: instance of Value which will be asssiggned to signal
    :param indexes: tuple on indexes where value should be updated
        in target array

    :return: function(value) -> tuple(valueHasChangedFlag, nextVal)
    """
    def updater(currentVal):
        if len(indexes) > 1:
            raise NotImplementedError("[TODO] implement for more indexes")

        _nextItemVal = nextItemVal.clone()
        if invalidate:
            _nextItemVal.vldMask = 0

        index = indexes[0]
        change = valueHasChanged(currentVal._getitem__val(index), _nextItemVal)
        currentVal._setitem__val(index, _nextItemVal)
        return (change, currentVal)

    return updater



from hwt.hdl.types.hdlType import HdlType


class String(HdlType):
    def __init__(self):
        super().__init__()

    def all_mask(self):
        return 1

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdl.types.stringVal import StringVal
            cls._valCls = StringVal
            return cls._valCls



from hwt.synthesizer.interfaceLevel.interfaceUtils.utils import NotSpecified
from hwt.synthesizer.interfaceLevel.mainBases import UnitBase
from hwt.synthesizer.interfaceLevel.unitImplHelpers import getRst, getClk


class InterfaceceImplDependentFns():
    """
    Interface functions which have high potential to be overloaded
    in concrete interface implementation
    """

    def _getIpCoreIntfClass(self):
        raise NotSpecified()

    def _initSimAgent(self):
        raise NotSpecified(("Override this function in your interface"
                            " implementation to have simultion agent"
                            " specified (%r)") % self)

    def _getAssociatedRst(self):
        """
        If interface has associated rst(_n) return it otherwise
        try to find rst(_n) on parent recursively
        """
        a = self._associatedRst

        if a is not None:
            return a

        p = self._parent
        assert p is not None

        if isinstance(p, UnitBase):
            return getRst(p)
        else:
            return p._getAssociatedRst()

    def _getAssociatedClk(self):
        """
        If interface has associated clk return it otherwise
        try to find clk on parent recursively
        """
        a = self._associatedClk

        if a is not None:
            return a

        p = self._parent
        assert p is not None

        if isinstance(p, UnitBase):
            return getClk(p)
        else:
            return p._getAssociatedClk()

    def _clone(self):
        """
        Create new instance of interface of same type and configuration
        """
        intf = self.__class__()
        intf._updateParamsFrom(self)
        return intf



import imp
import importlib
import inspect
import os
import sys

from hwt.hdl.constants import Time
from hwt.serializer.simModel.serializer import SimModelSerializer
from hwt.simulator.agentConnector import autoAddAgents
from hwt.simulator.hdlSimulator import HdlSimulator
from hwt.simulator.simModel import SimModel
from hwt.simulator.simSignal import SimSignal
from hwt.simulator.simSignalProxy import IndexSimSignalProxy
from hwt.simulator.types.simBits import simBitsT
from hwt.simulator.vcdHdlSimConfig import VcdHdlSimConfig
from hwt.synthesizer.interfaceLevel.interfaceUtils.proxy import InterfaceProxy
from hwt.synthesizer.interfaceLevel.mainBases import InterfaceBase
from hwt.synthesizer.unit import Unit
from hwt.synthesizer.utils import toRtl
from typing import Optional


def simPrepare(unit: Unit, modelCls: Optional[SimModel]=None,
               targetPlatform=None,
               dumpModelIn: str=None, onAfterToRtl=None):
    """
    Create simulation model and connect it with interfaces of original unit
    and decorate it with agents

    :param unit: interface level unit which you wont prepare for simulation
    :param modelCls: class of rtl simulation model to run simulation on,
        if is None rtl sim model will be generated from unit
    :param targetPlatform: target platform for this synthes
    :param dumpModelIn: folder to where put sim model files
        (if is None sim model will be constructed only in memory)
    :param onAfterToRtl: callback fn(unit, modelCls) which will be called
        after unit will be synthesised to rtl

    :return: tuple (fully loaded unit with connected sim model,
        connected simulation model,
        simulation processes of agents
        )
    """
    if modelCls is None:
        modelCls = toSimModel(
            unit, targetPlatform=targetPlatform, dumpModelIn=dumpModelIn)
    else:
        # to instantiate hierarchy of unit
        toSimModel(unit)

    if onAfterToRtl:
        onAfterToRtl(unit, modelCls)

    reconnectUnitSignalsToModel(unit, modelCls)
    model = modelCls()
    procs = autoAddAgents(unit)
    return unit, model, procs


def toSimModel(unit, targetPlatform=None, dumpModelIn=None):
    """
    Create a simulation model for unit

    :param unit: interface level unit which you wont prepare for simulation
    :param targetPlatform: target platform for this synthes
    :param dumpModelIn: folder to where put sim model files
        (otherwise sim model will be constructed only in memory)
    """
    sim_code = toRtl(unit,
                     targetPlatform=targetPlatform,
                     saveTo=dumpModelIn,
                     serializer=SimModelSerializer)
    if dumpModelIn is not None:
        d = os.path.join(os.getcwd(), dumpModelIn)
        dInPath = d in sys.path
        if not dInPath:
            sys.path.insert(0, d)
        if unit._name in sys.modules:
            del sys.modules[unit._name]
        simModule = importlib.import_module(unit._name)

        if not dInPath:
            sys.path.remove(d)
    else:
        simModule = imp.new_module('simModule')
        # python supports only ~100 opened brackets
        # it exceded it throws MemoryError: s_push: parser stack overflow
        exec(sim_code, simModule.__dict__)

    return simModule.__dict__[unit._name]


def reconnectUnitSignalsToModel(synthesisedUnitOrIntf, modelCls,
                                destroyProxies=False):
    """
    Reconnect model signals to unit to run simulation with simulation model
    but use original unit interfaces for communication

    :param synthesisedUnitOrIntf: interface where should be signals
        replaced from signals from modelCls
    :param modelCls: simulation model form where signals
        for synthesisedUnitOrIntf should be taken
    :param destroyProxies: destroy proxies, is true when this interface
        is part of array and potentially proxies under this
        interface would interfere with other proxies
    """
    obj = synthesisedUnitOrIntf
    subInterfaces = obj._interfaces
    isProxy = isinstance(obj, InterfaceProxy)
    hasProxies = isinstance(obj, InterfaceBase) and bool(obj._arrayElemCache)

    if destroyProxies:
        assert not isProxy, "Proxy should be already destroyed"

    if not isProxy and subInterfaces:
        for intf in subInterfaces:
            # proxies are destroyed on original interfaces and only proxies on
            # array items will remain
            reconnectUnitSignalsToModel(
                intf, modelCls, destroyProxies=destroyProxies or hasProxies)

        if not destroyProxies and hasProxies:
            # if this this interface has proxies for array items let them
            # reconnect
            for proxy in obj._arrayElemCache:
                reconnectUnitSignalsToModel(proxy, modelCls)
    else:
        if isProxy:
            if hasProxies:
                # this obj will become only container of elements for array
                for subIntf in subInterfaces:
                    # delete attributes because we can not use them
                    # in simulation because they are managed by children
                    delattr(obj, subIntf._name)
                del obj._interfaces

                # if this interface is array we have to replace signals in
                # array items
                for item in obj._arrayElemCache:
                    reconnectUnitSignalsToModel(item, modelCls)
            else:
                if subInterfaces:
                    # let children reconnect
                    for intf in subInterfaces:
                        reconnectUnitSignalsToModel(intf, modelCls)
                else:
                    assert obj._itemsInOne == 1, (
                        obj, "Now there should be proxies only for leaves"
                             "and proxies on partial arrays should be deleted")

                    # setup proxy on signal from model
                    p = obj._origIntf
                    while isinstance(p, InterfaceProxy):
                        assert p._itemsInOne == 1, (
                            p,
                            "Now there should be proxies only for leaves"
                            "and proxies on partial arrays should be deleted")
                        p = p._origIntf

                    s = p._sigInside
                    index = obj._getMySigSelector()

                    try:
                        upperIndex, lowerIndex = index
                        width = upperIndex - lowerIndex
                    except TypeError:
                        lowerIndex = None
                        upperIndex = index
                        width = 1
                    obj._sigInside = IndexSimSignalProxy(obj._origIntf._name,
                                                         p._sigInside,
                                                         simBitsT(
                                                             width,
                                                             s._dtype.signed),
                                                         upperIndex,
                                                         lowerIndex)
        else:
            # reconnect signal from model
            s = synthesisedUnitOrIntf
            s._sigInside = getattr(modelCls, s._sigInside.name)


def simUnitVcd(simModel, stimulFunctions, outputFile=sys.stdout,
               time=100 * Time.ns):
    """
    Syntax sugar
    If outputFile is string try to open it as file

    :return: hdl simulator object
    """
    assert isinstance(simModel, SimModel), \
        "Class of SimModel is required (got %r)" % (simModel)
    if isinstance(outputFile, str):
        d = os.path.dirname(outputFile)
        if d:
            os.makedirs(d, exist_ok=True)
        with open(outputFile, 'w') as f:
            return _simUnitVcd(simModel, stimulFunctions,
                               outputFile=f, time=time)
    else:
        return _simUnitVcd(simModel, stimulFunctions,
                           outputFile=outputFile, time=time)


def _simUnitVcd(simModel, stimulFunctions, outputFile, time):
    """
    :param unit: interface level unit to simulate
    :param stimulFunctions: iterable of function with single param env
        (simpy environment) which are driving the simulation
    :param outputFile: file where vcd will be dumped
    :param time: endtime of simulation, time units are defined in HdlSimulator
    :return: hdl simulator object
    """
    sim = HdlSimulator()

    # configure simulator to log in vcd
    sim.config = VcdHdlSimConfig(outputFile)

    # run simulation, stimul processes are register after initial
    # initialization
    sim.simUnit(simModel, time=time, extraProcesses=stimulFunctions)
    return sim


class CallbackLoop(object):
    def __init__(self, sig: SimSignal, fn, shouldBeEnabledFn):
        """
        :param sig: signal on which write callback should be used
        :attention: if condFn is None callback function is always executed

        :ivra fn: function/generator which is callback which should be executed
        :ivar isGenerator: flag if callback function is generator
            or normal function
        :ivar _callbackIndex: index of callback in write callbacks on sig,
            if is None callback was not registered yet
        :ivar shouldBeEnabledFn; function() -> bool, which returns True if this
            callback loop should be enabled
        """
        assert not isinstance(fn, CallbackLoop)
        self.fn = fn
        self.isGenerator = inspect.isgeneratorfunction(fn)
        self.shouldBeEnabledFn = shouldBeEnabledFn
        self._callbackIndex = None

        try:
            # if sig is interface we need internal signal
            self.sig = sig._sigInside
        except AttributeError:
            self.sig = sig

    def setEnable(self, en, sim):
        if self._callbackIndex is None:
            return

        if en:
            c = self.onWriteCallback
        else:
            c = None
        self.sig._writeCallbacks[self._callbackIndex] = c

    def onWriteCallback(self, sim):
        if self.isGenerator:
            yield from self.fn(sim)
        else:
            self.fn(sim)

    def __call__(self, sim):
        """
        Process for injecting of this callback loop into simulator
        """
        self._callbackIndex = self.sig.registerWriteCallback(
            self.onWriteCallback,
            self.shouldBeEnabledFn)
        return
        yield


class OnRisingCallbackLoop(CallbackLoop):

    def onWriteCallback(self, sim):
        if bool(sim.read(self.sig)._onRisingEdge__val(sim.now)):
            if self.isGenerator:
                yield from self.fn(sim)
            else:
                self.fn(sim)


class OnFallingCallbackLoop(CallbackLoop):

    def onWriteCallback(self, sim):
        if bool(sim.read(self.sig)._onFallingEdge__val(sim.now)):
            if self.isGenerator:
                yield from self.fn(sim)
            else:
                self.fn(sim)


def oscilate(sig, period=10 * Time.ns, initWait=0):
    """
    Oscillative simulation driver for your signal
    (usually used as clk generator)
    """
    def oscillateStimul(s):
        s.write(False, sig)
        halfPeriod = period / 2
        yield s.wait(initWait)

        while True:
            yield s.wait(halfPeriod)
            s.write(True, sig)
            yield s.wait(halfPeriod)
            s.write(False, sig)

    return oscillateStimul



from hwt.hdl.types.hdlType import HdlType


class HEnum(HdlType):
    """
    Hdl enum type

    :ivar name: name of this type
    :ivar _allValues: tuple of all values for this enum
    :note: for each value there is a property on this type object
    """

    def __init__(self, name, valueNames):
        """
        :param name: name for this type
        :param valueNames: sequence of string which will be used as names
            for enum members
        """
        super(HEnum, self).__init__()
        self.name = name
        self._allValues = tuple(valueNames)
        for name in valueNames:
            v = self.fromPy(name)
            assert not hasattr(self, name)
            setattr(self, name, v)

    def all_mask(self):
        return 1

    def bit_length(self):
        return len(self._allValues).bit_length()

    @classmethod
    def getValueCls(cls):
        try:
            return cls._valCls
        except AttributeError:
            from hwt.hdl.types.enumVal import HEnumVal
            cls._valCls = HEnumVal
            return cls._valCls



from copy import copy


class SerializerCtx():
    """
    Serializer context

    :ivar scope: instance of NameScope used to check id availability
    :ivar indent: number of visual indentations for code in this context
    :ivar createTmpVarFn: function (sugestedName, dtype) returns variable
        this function will be called to create tmp variables
    :ivar constCache: constant cache to extract frequently used large constant
        values from code (visual improvement)
    """

    def __init__(self, scope, indent, createTmpVarFn, constCache=None):
        self.scope = scope
        self.indent = indent

        if createTmpVarFn is None:
            self.createTmpVarFn = self.defaultCreateTmpVarFn
        else:
            self.createTmpVarFn = createTmpVarFn

        self.constCache = constCache

    def defaultCreateTmpVarFn(self, sugestedName, dtype):
        raise NotImplementedError()

    def withIndent(self, indent=1):
        """
        Create copy of this context with increased indent
        """
        ctx = copy(self)
        ctx.indent += indent
        return ctx



import fnmatch
import os


def find_files(directory, pattern, recursive=True):
    """
    Find files by pattern in directory
    """
    if not os.path.isdir(directory):
        if os.path.exists(directory):
            raise IOError(directory + ' is not directory')
        else:
            raise IOError(directory + " does not exists")
    if recursive:
        for root, _, files in os.walk(directory):
            for basename in files:
                if fnmatch.fnmatch(basename, pattern):
                    filename = os.path.join(root, basename)
                    yield filename
    else:
        root = directory
        for basename in os.listdir(root):
            if fnmatch.fnmatch(basename, pattern):
                filename = os.path.join(root, basename)
                if os.path.isfile(filename):
                    yield filename



from hwt.simulator.agentBase import AgentWitReset
from hwt.interfaces.agents.signal import DEFAULT_CLOCK
from hwt.hdl.constants import NOP
from collections import deque


def toGenerator(fn):
    def asGen(sim):
        fn(sim)
        return
        yield

    return asGen


class TristateAgent(AgentWitReset):
    def __init__(self, intf, allowNoReset=True):
        super(TristateAgent, self).__init__(intf, allowNoReset=allowNoReset)
        self.data = deque()
        # can be (1: pull-up, 0: pull-down, None: disconnected)
        self.pullMode = 1
        self.selfSynchronization = True
        self.collectData = True

    def monitor(self, sim):
        intf = self.intf
        # read in pre-clock-edge
        t = sim.read(intf.t)
        o = sim.read(intf.o)

        if self.pullMode is not None and sim.now > 0:
            assert t.vldMask, (
                sim.now, intf, "This mode, this value => ioblock would burn")
            assert o.vldMask, (
                sim.now, intf, "This mode, this value => ioblock would burn")
            assert self.pullMode != o.val, (
                sim.now, intf, "This mode, this value => ioblock would burn")

        if t.val:
            v = o
        else:
            v = self.pullMode

        sim.write(v, intf.i)
        if self.collectData and sim.now > 0 and self.notReset(sim):
            self.data.append(v)

    def getMonitors(self):
        return [self.onTWriteCallback__init]

    def onTWriteCallback(self, sim):
        self.monitor(sim)
        return
        yield

    def _write(self, val, sim):
        if val is NOP:
            # controll now has slave
            t = 0
            o = self.pullMode
        else:
            # controll now has this agent
            t = 1
            o = val

        intf = self.intf
        w = sim.write
        w(t, intf.t)
        w(o, intf.o)

    def onTWriteCallback__init(self, sim):
        """
        Process for injecting of this callback loop into simulator
        """
        yield from self.onTWriteCallback(sim)
        self.intf.t._sigInside.registerWriteCallback(
            self.onTWriteCallback,
            self.getEnable)
        self.intf.o._sigInside.registerWriteCallback(
            self.onTWriteCallback,
            self.getEnable)

    def driver(self, sim):
        while True:
            if self.data:
                b = self.data.popleft()
                if b == self.START:
                    return
                self.sda._write(b, sim)
            if self.selfSynchronization:
                yield sim.wait(DEFAULT_CLOCK)
            else:
                break


class TristateClkAgent(TristateAgent):
    def __init__(self, intf, onRisingCallback=None, onFallingCallback=None):
        super(TristateClkAgent, self).__init__(intf)
        self.onRisingCallback = onRisingCallback
        self.onFallingCallback = onFallingCallback
        self.period = DEFAULT_CLOCK

    def driver(self, sim):
        o = self.intf.o
        high = self.pullMode
        low = not self.pullMode
        halfPeriod = self.period / 2

        sim.write(low, o)
        sim.write(1, self.intf.t)
        if high:
            onHigh = self.onRisingCallback
            onLow = self.onFallingCallback
        else:
            onHigh = self.onFallingCallback
            onLow = self.onRisingCallback

        while True:
            yield sim.wait(halfPeriod)
            sim.write(high, o)

            if onHigh:
                sim.process(onHigh(sim))

            yield sim.wait(halfPeriod)
            sim.write(low, o)

            if onLow:
                sim.process(onLow(sim))

    def monitor(self, sim):
        intf = self.intf
        yield sim.waitOnCombUpdate()
        # read in pre-clock-edge
        t = sim.read(intf.t)
        o = sim.read(intf.o)

        if sim.now > 0 and self.pullMode is not None:
            assert t.vldMask, (
                sim.now, intf, "This mode, this value => ioblock would burn")
            assert o.vldMask, (
                sim.now, intf, "This mode, this value => ioblock would burn")
            assert self.pullMode != o.val, (
                sim.now, intf, "This mode, this value => ioblock would burn")

        if t.val:
            v = o
        else:
            v = self.pullMode

        last = sim.read(intf.i)
        sim.write(v, intf.i)

        if self.onRisingCallback and (not last.val or not last.vldMask) and v:
            sim.process(self.onRisingCallback(sim))

        if self.onFallingCallback and not v and (last.val or not last.vldMask):
            sim.process(self.onFallingCallback(sim))

    def getMonitors(self):
        return [self.onTWriteCallback__init]

    def onTWriteCallback(self, sim):
        yield from self.monitor(sim)
        return
        yield

    def onTWriteCallback__init(self, sim):
        """
        Process for injecting of this callback loop into simulator
        """
        yield from self.onTWriteCallback(sim)
        self.intf.t._sigInside.registerWriteCallback(
            self.onTWriteCallback,
            self.getEnable)
        self.intf.o._sigInside.registerWriteCallback(
            self.onTWriteCallback,
            self.getEnable)




class AndReducedContainer(list):
    """
    Container of and terms usually used as container for condition
    """

    def __init__(self):
        super(AndReducedContainer, self).__init__()
        self.__s = set()

    def add(self, item):
        if item in self.__s:
            return False
        else:
            self.__s.add(item)
            list.append(self, item)
            return True

    def update(self, other):
        assert isinstance(other, AndReducedContainer)
        for i in other:
            self.add(i)

    def __contains__(self, key):
        return key in self.__s



from hwt.hdl.portItem import PortItem
from hwt.hdl.constants import DIRECTION
from hwt.synthesizer.rtlLevel.signalUtils.walkers import signalHasDriver


def portItemfromSignal(s, entity):
    if signalHasDriver(s):
        d = DIRECTION.OUT
    else:
        d = DIRECTION.IN
    pi = PortItem(s.name, d, s._dtype, entity)
    if not hasattr(s, '_interface'):
        from hwt.interfaces.std import Signal
        t = s._dtype
        s._interface = Signal(dtype=t)
    pi._interface = s._interface

    return pi



from hwt.hdl.hdlObject import HdlObject


class Architecture(HdlObject):
    """
    Hdl container of internal structure of unit
    """

    def __init__(self, entity):
        self.entity = entity
        self.name = "rtl"
        self.variables = []
        self.processes = []
        self.components = []
        self.componentInstances = []

    def getEntityName(self):
        return self.entity.name
